import json
import os
import threading
from datetime import datetime

from sqlmodel import Session, select

from pixlvault.database import DBPriority
from pixlvault.db_models.picture import Picture
from pixlvault.utils.image_processing.image_utils import ImageUtils
from pixlvault.pixl_logging import get_logger
from pixlvault.stacking import (
    assign_picture_to_stack,
    get_or_create_stack_for_picture,
    parse_stack_tags_from_filename,
)
from pixlvault.tasks.base_task import BaseTask


logger = get_logger(__name__)

_CONFIG_LOCK = threading.Lock()


def load_watch_folders(config_path: str) -> list[dict]:
    """Load the watch_folders list from the server config file.

    Each entry in the returned list is a dict with the following fields:

        folder (str): Absolute path to the directory to monitor recursively.
        delete_after_import (bool): When True, source files are deleted from
            the watch folder after a successful import. Defaults to False.
        last_checked (float): Unix timestamp of the last scan. Managed
            internally by the finder; there is no need to set this manually.

    Args:
        config_path: Path to the server-config.json file.

    Returns:
        List of watch folder entry dicts, or an empty list on error.
    """
    if not config_path or not os.path.exists(config_path):
        return []
    with _CONFIG_LOCK:
        try:
            with open(config_path, "r") as handle:
                config = json.load(handle)
            return list(config.get("watch_folders", []) or [])
        except Exception as exc:
            logger.error("Failed to read watch_folders: %s", exc)
            return []


def persist_watch_folders(config_path: str, watch_folders: list[dict]):
    if not config_path:
        return
    with _CONFIG_LOCK:
        try:
            config = {}
            if os.path.exists(config_path):
                with open(config_path, "r") as handle:
                    config = json.load(handle)
            config["watch_folders"] = watch_folders
            with open(config_path, "w") as handle:
                json.dump(config, handle, indent=2)
        except Exception as exc:
            logger.error("Failed to persist watch_folders: %s", exc)


class WatchFolderImportTask(BaseTask):
    """Task that imports discovered files from watch folders."""

    def __init__(
        self,
        database,
        config_path: str,
        candidate_files: list[dict],
        updated_watch_folders: list[dict],
        total_candidates: int,
    ):
        super().__init__(
            task_type="WatchFolderImportTask",
            params={
                "candidate_count": len(candidate_files or []),
                "total_candidates": int(total_candidates or 0),
            },
        )
        self._db = database
        self._config_path = config_path
        self._candidate_files = candidate_files or []
        self._updated_watch_folders = updated_watch_folders or []
        self._total_candidates = int(total_candidates or 0)

    def _run_task(self):
        new_pictures = []
        stack_assignments = []
        delete_paths = []

        for candidate in self._candidate_files:
            file_path = candidate.get("file_path")
            if not file_path:
                continue

            try:
                pixel_sha = ImageUtils.calculate_hash_from_file_path(file_path)
            except Exception as exc:
                logger.warning("Failed to hash watched file %s: %s", file_path, exc)
                continue

            def find_existing(session: Session, hash_value: str):
                return session.exec(
                    select(Picture).where(Picture.pixel_sha == hash_value)
                ).first()

            existing = self._db.run_task(find_existing, pixel_sha)
            if existing:
                logger.debug("Already have picture with sha %s, skipping", pixel_sha)
                continue

            try:
                pic = ImageUtils.create_picture_from_file(
                    image_root_path=self._db.image_root,
                    source_file_path=file_path,
                    pixel_sha=pixel_sha,
                )
                pic.imported_at = datetime.now()
                new_pictures.append(pic)

                stack_id, source_id = parse_stack_tags_from_filename(file_path)
                if stack_id or source_id:
                    stack_assignments.append((pic, stack_id, source_id))
                if bool(candidate.get("delete_after_import", False)):
                    delete_paths.append(file_path)
            except Exception as exc:
                logger.warning("Failed to import watched file %s: %s", file_path, exc)

        changed = []
        imported_ids = []
        if new_pictures:

            def insert_pictures(session: Session, pictures: list[Picture]):
                session.add_all(pictures)
                session.commit()
                for pic in pictures:
                    session.refresh(pic)
                return pictures

            inserted = self._db.run_task(
                insert_pictures,
                new_pictures,
                priority=DBPriority.IMMEDIATE,
            )

            if stack_assignments:

                def apply_stack_assignments(session: Session, assignments: list[tuple]):
                    for pic, stack_id, source_id in assignments:
                        if pic.id is None:
                            continue
                        if stack_id:
                            assign_picture_to_stack(session, pic.id, stack_id)
                            continue
                        if source_id:
                            resolved_stack_id = get_or_create_stack_for_picture(
                                session, source_id
                            )
                            if resolved_stack_id:
                                assign_picture_to_stack(
                                    session,
                                    pic.id,
                                    resolved_stack_id,
                                )

                self._db.run_task(
                    apply_stack_assignments,
                    stack_assignments,
                    priority=DBPriority.IMMEDIATE,
                )

            for pic in inserted or []:
                if getattr(pic, "id", None) is None:
                    continue
                imported_ids.append(pic.id)
                changed.append((Picture, pic.id, "imported_at", pic.imported_at))

            logger.info("Added %d new pictures from watch folders.", len(imported_ids))

            if delete_paths:
                for file_path in delete_paths:
                    try:
                        os.remove(file_path)
                    except Exception as exc:
                        logger.warning(
                            "Failed to delete watched file %s: %s",
                            file_path,
                            exc,
                        )

        if self._updated_watch_folders:
            persist_watch_folders(self._config_path, self._updated_watch_folders)

        return {
            "changed_count": len(changed),
            "changed": changed,
            "imported_picture_ids": imported_ids,
            "candidate_count": self._total_candidates,
        }



from __future__ import annotations

import os
import uuid
from datetime import datetime
from io import BytesIO
from typing import Any

from PIL import Image
from sqlmodel import Session, select

from pixlvault.db_models import Picture, PictureSetMember, PictureStack
from pixlvault.image_plugins.base import ImagePlugin
from pixlvault.utils.image_processing.image_utils import ImageUtils
from pixlvault.pixl_logging import get_logger
from pixlvault.stacking import get_or_create_stack_for_picture

logger = get_logger(__name__)

_VIDEO_FORMATS = {"MP4", "WEBM", "MOV", "AVI", "MKV"}


def _load_input_images(
    server,
    picture_ids: list[int],
) -> list[tuple[Picture, Image.Image, str, str]]:
    def fetch_pictures(session: Session, ids: list[int]):
        return session.exec(select(Picture).where(Picture.id.in_(ids))).all()

    pictures = server.vault.db.run_task(fetch_pictures, picture_ids)
    picture_map = {pic.id: pic for pic in pictures if pic.id is not None}

    loaded: list[tuple[Picture, Image.Image, str, str]] = []
    for picture_id in picture_ids:
        pic = picture_map.get(picture_id)
        if pic is None:
            raise ValueError(f"Picture not found: {picture_id}")
        if not pic.file_path:
            raise ValueError(f"Picture missing file path: {picture_id}")
        resolved_path = ImageUtils.resolve_picture_path(
            server.vault.image_root, pic.file_path
        )
        if not resolved_path or not os.path.isfile(resolved_path):
            raise ValueError(f"Picture file missing: {picture_id}")
        frame = ImageUtils.load_image_or_video(resolved_path)
        if frame is None:
            raise ValueError(f"Could not load image/video data: {picture_id}")
        try:
            frame_image = Image.fromarray(frame).convert("RGB")
        except Exception as exc:
            raise ValueError(
                f"Could not convert image/video data to PIL image: {picture_id}"
            ) from exc
        source_format = str(pic.format or "").strip().upper() or "PNG"
        loaded.append((pic, frame_image, source_format, resolved_path))
    return loaded


def _save_output_images(image: Any, source_format: str) -> tuple[bytes, str]:
    normalized = (source_format or "PNG").upper()

    if (
        isinstance(image, tuple)
        and len(image) == 2
        and isinstance(image[0], (bytes, bytearray))
        and isinstance(image[1], str)
    ):
        ext = image[1] if image[1].startswith(".") else f".{image[1]}"
        return bytes(image[0]), ext

    if normalized in _VIDEO_FORMATS:
        ext = f".{normalized.lower()}"
        if isinstance(image, (bytes, bytearray)):
            return bytes(image), ext
        if not isinstance(image, Image.Image):
            raise ValueError(
                "Plugin output for video sources must be PIL image or encoded bytes"
            )
        normalized = "PNG"

    if isinstance(image, (bytes, bytearray)):
        if normalized in {"JPG", "JPEG"}:
            ext = ".jpg"
        elif normalized == "WEBP":
            ext = ".webp"
        elif normalized == "BMP":
            ext = ".bmp"
        elif normalized in {"TIFF", "TIF"}:
            ext = ".tiff"
        else:
            ext = ".png"
        return bytes(image), ext

    if not isinstance(image, Image.Image):
        raise ValueError("Plugin output must be PIL image or encoded bytes")

    if normalized in {"JPG", "JPEG"}:
        ext = ".jpg"
        save_format = "JPEG"
    elif normalized in {"WEBP"}:
        ext = ".webp"
        save_format = "WEBP"
    elif normalized in {"BMP"}:
        ext = ".bmp"
        save_format = "BMP"
    elif normalized in {"TIFF", "TIF"}:
        ext = ".tiff"
        save_format = "TIFF"
    else:
        ext = ".png"
        save_format = "PNG"

    out = image.convert("RGB")
    buf = BytesIO()
    if save_format == "JPEG":
        out.save(buf, format=save_format, quality=95)
    else:
        out.save(buf, format=save_format)
    return buf.getvalue(), ext


def _import_output_images(
    server,
    output_entries: list[tuple[bytes, str]],
) -> tuple[list[int], list[int], list[int]]:
    if not output_entries:
        return [], [], []

    shas = [
        ImageUtils.calculate_hash_from_bytes(image_bytes)
        for image_bytes, _ in output_entries
    ]

    existing = server.vault.db.run_immediate_read_task(
        lambda session: Picture.find(session, pixel_shas=shas, include_unimported=True)
    )
    existing_map = {pic.pixel_sha: pic for pic in existing}

    new_entries = [
        (entry, sha)
        for entry, sha in zip(output_entries, shas)
        if sha not in existing_map
    ]

    new_pictures = []
    for (img_bytes, ext), sha in new_entries:
        picture_uuid = f"{uuid.uuid4()}{ext}"
        new_pictures.append(
            ImageUtils.create_picture_from_bytes(
                image_root_path=server.vault.image_root,
                image_bytes=img_bytes,
                picture_uuid=picture_uuid,
                pixel_sha=sha,
            )
        )

    def persist(session: Session):
        if not new_pictures:
            return []
        session.add_all(new_pictures)
        session.commit()
        for pic in new_pictures:
            session.refresh(pic)
        return new_pictures

    if new_pictures:
        new_pictures = server.vault.db.run_task(persist)

        def mark_imported(session: Session, ids: list[int]):
            now = datetime.utcnow()
            pictures = session.exec(select(Picture).where(Picture.id.in_(ids))).all()
            for pic in pictures:
                if pic.imported_at is None:
                    pic.imported_at = now
                    session.add(pic)
            session.commit()

        server.vault.db.run_task(
            mark_imported,
            [pic.id for pic in new_pictures if pic.id is not None],
        )

    new_ids = [pic.id for pic in new_pictures if pic.id is not None]
    duplicate_ids = [
        pic.id
        for sha in shas
        if (pic := existing_map.get(sha)) is not None and pic.id is not None
    ]

    new_map: dict[str, int] = {}
    for (_entry, sha), pic in zip(new_entries, new_pictures):
        if pic.id is not None:
            new_map[sha] = pic.id

    ordered_output_ids: list[int] = []
    for sha in shas:
        if sha in new_map:
            ordered_output_ids.append(new_map[sha])
            continue
        existing_pic = existing_map.get(sha)
        if existing_pic is not None and existing_pic.id is not None:
            ordered_output_ids.append(existing_pic.id)

    return new_ids, duplicate_ids, ordered_output_ids


def _assign_outputs_to_stack_top(server, stack_id: int, picture_ids: list[int]) -> None:
    if not stack_id or not picture_ids:
        return

    def update_stack(session: Session):
        stack = session.get(PictureStack, stack_id)
        if stack is None:
            return
        pics = session.exec(select(Picture).where(Picture.stack_id == stack_id)).all()
        has_positions = any(pic.stack_position is not None for pic in pics)
        shift = len(picture_ids)
        if has_positions and shift:
            for pic in pics:
                if pic.id in picture_ids:
                    continue
                if pic.stack_position is not None:
                    pic.stack_position += shift
                    session.add(pic)

        for idx, pic_id in enumerate(picture_ids):
            pic = session.get(Picture, pic_id)
            if pic is None:
                continue
            pic.stack_id = stack_id
            pic.stack_position = idx
            session.add(pic)

        stack.updated_at = datetime.utcnow()
        session.add(stack)
        session.commit()

    server.vault.db.run_task(update_stack)


def _propagate_output_picture_sets(
    server,
    source_picture_ids: list[int],
    output_picture_ids: list[int],
) -> None:
    if not source_picture_ids or not output_picture_ids:
        return

    source_to_outputs: dict[int, set[int]] = {}
    for source_id, output_id in zip(source_picture_ids, output_picture_ids):
        if source_id is None or output_id is None:
            continue
        if source_id == output_id:
            continue
        source_to_outputs.setdefault(int(source_id), set()).add(int(output_id))

    if not source_to_outputs:
        return

    source_ids = list(source_to_outputs.keys())
    output_ids = sorted({oid for ids in source_to_outputs.values() for oid in ids})

    def copy_memberships(session: Session):
        source_memberships = session.exec(
            select(PictureSetMember).where(PictureSetMember.picture_id.in_(source_ids))
        ).all()

        source_set_ids: dict[int, set[int]] = {}
        for member in source_memberships:
            source_set_ids.setdefault(int(member.picture_id), set()).add(
                int(member.set_id)
            )

        desired_pairs: set[tuple[int, int]] = set()
        for source_id, out_ids in source_to_outputs.items():
            set_ids = source_set_ids.get(source_id)
            if not set_ids:
                continue
            for out_id in out_ids:
                for set_id in set_ids:
                    desired_pairs.add((set_id, out_id))

        if not desired_pairs:
            return

        desired_set_ids = sorted({set_id for set_id, _ in desired_pairs})
        existing = session.exec(
            select(PictureSetMember).where(
                PictureSetMember.picture_id.in_(output_ids),
                PictureSetMember.set_id.in_(desired_set_ids),
            )
        ).all()
        existing_pairs = {
            (int(member.set_id), int(member.picture_id)) for member in existing
        }

        inserts = [
            PictureSetMember(set_id=set_id, picture_id=picture_id)
            for set_id, picture_id in sorted(desired_pairs - existing_pairs)
        ]
        if inserts:
            session.add_all(inserts)
            session.commit()

    server.vault.db.run_task(copy_memberships)


def apply_plugin_to_pictures(
    server,
    plugin: ImagePlugin,
    picture_ids: list[int],
    parameters: dict[str, Any] | None,
    progress_reporter=None,
    error_reporter=None,
) -> dict[str, Any]:
    loaded = _load_input_images(server, picture_ids)

    progress_events: list[dict[str, Any]] = []
    errors: list[dict[str, Any]] = []

    params = parameters or {}

    def progress_cb(payload):
        progress_events.append(payload)
        if progress_reporter is not None:
            progress_reporter(payload)

    def error_cb(payload):
        errors.append(payload)
        if error_reporter is not None:
            error_reporter(payload)

    outputs: list[Any] = [None] * len(loaded)
    image_indices: list[int] = []
    image_inputs: list[Image.Image] = []

    for idx, (_pic, pil_image, source_format, source_path) in enumerate(loaded):
        if source_format in _VIDEO_FORMATS and plugin.supports_videos:
            if type(plugin).run_video is not ImagePlugin.run_video:
                outputs[idx] = plugin.run_video(
                    source_path,
                    parameters=params,
                    progress_callback=progress_cb,
                    error_callback=error_cb,
                )
            else:
                image_indices.append(idx)
                image_inputs.append(pil_image)
        else:
            image_indices.append(idx)
            image_inputs.append(pil_image)

    if image_inputs:
        image_outputs = plugin.run(
            image_inputs,
            parameters=params,
            progress_callback=progress_cb,
            error_callback=error_cb,
        )
        if len(image_outputs) != len(image_inputs):
            raise ValueError(
                f"Plugin '{plugin.name}' returned {len(image_outputs)} images for {len(image_inputs)} inputs"
            )
        for out_idx, loaded_idx in enumerate(image_indices):
            outputs[loaded_idx] = image_outputs[out_idx]

    if any(output is None for output in outputs):
        raise ValueError(
            f"Plugin '{plugin.name}' did not return outputs for all inputs"
        )

    output_entries: list[tuple[bytes, str]] = []
    source_picture_ids: list[int] = []
    for idx, output in enumerate(outputs):
        pic, _, source_format, _source_path = loaded[idx]
        output_bytes, ext = _save_output_images(output, source_format)
        output_entries.append((output_bytes, ext))
        source_picture_ids.append(pic.id)

    new_ids, duplicate_ids, ordered_output_ids = _import_output_images(
        server, output_entries
    )

    _propagate_output_picture_sets(server, source_picture_ids, ordered_output_ids)

    for source_id, out_id in zip(source_picture_ids, ordered_output_ids):
        stack_id = server.vault.db.run_task(get_or_create_stack_for_picture, source_id)
        if stack_id:
            _assign_outputs_to_stack_top(server, stack_id, [out_id])

    return {
        "plugin": plugin.name,
        "picture_ids": picture_ids,
        "created_picture_ids": new_ids,
        "duplicate_picture_ids": duplicate_ids,
        "output_picture_ids": ordered_output_ids,
        "progress": progress_events,
        "errors": errors,
    }



"""Caption and tag processing utilities for picture export."""

from pixlvault.db_models.tag import TAG_EMPTY_SENTINEL


class CaptionUtils:
    """Utility methods for building caption and tag strings from pictures."""

    @staticmethod
    def _build_tag_caption(picture) -> str:
        """Build a comma-separated tag string from a picture's tags."""
        tags = []
        for tag in getattr(picture, "tags", []) or []:
            tag_value = getattr(tag, "tag", None)
            if tag_value in (None, TAG_EMPTY_SENTINEL):
                continue
            tags.append(tag_value)
        return ", ".join(tags)

    @staticmethod
    def _build_character_caption(picture) -> str:
        """Build a comma-separated character name string from a picture's characters."""
        character_names = []
        for character in getattr(picture, "characters", []) or []:
            name_value = getattr(character, "name", None)
            if name_value:
                character_names.append(name_value)
        return ", ".join(character_names)



"""Image loading, thumbnail generation, metadata extraction, and picture creation utilities."""

import cv2
import hashlib
import os
import piexif
import tempfile
import uuid

from datetime import datetime, timezone
from fractions import Fraction
from io import BytesIO
from typing import Optional

import numpy as np
from PIL import ExifTags, Image, ImageOps

try:
    from PIL.TiffImagePlugin import IFDRational
except Exception:  # pragma: no cover - optional import
    IFDRational = None

from pixlvault.pixl_logging import get_logger
from pixlvault.db_models.picture import Picture
from pixlvault.utils.image_processing.video_utils import VideoUtils

logger = get_logger(__name__)

THUMBNAIL_FORMAT = "WEBP"
THUMBNAIL_EXTENSION = ".webp"
THUMBNAIL_QUALITY = 80
THUMBNAIL_WEBP_METHOD = 2


class ImageUtils:
    """Utility methods for image loading, thumbnails, metadata, and picture creation."""

    @staticmethod
    def _coerce_metadata_value(value):
        """Coerce a raw metadata value to a JSON-serialisable Python type."""
        if IFDRational is not None and isinstance(value, IFDRational):
            try:
                return float(value)
            except Exception:
                return str(value)
        if isinstance(value, Fraction):
            try:
                return float(value)
            except Exception:
                return str(value)
        if hasattr(value, "numerator") and hasattr(value, "denominator"):
            try:
                return float(value)
            except Exception:
                return str(value)
        if isinstance(value, bytes):
            try:
                return value.decode("utf-8", errors="replace")
            except Exception:
                return repr(value)
        if isinstance(value, (list, tuple)):
            return [ImageUtils._coerce_metadata_value(v) for v in value]
        if isinstance(value, dict):
            return {
                str(k): ImageUtils._coerce_metadata_value(v) for k, v in value.items()
            }
        if isinstance(value, datetime):
            return value.isoformat()
        return value

    @staticmethod
    def extract_embedded_metadata(file_path: str) -> dict:
        """Extract embedded EXIF and PNG metadata from an image file."""
        if not file_path or not os.path.exists(file_path):
            return {}
        metadata = {}
        try:
            with Image.open(file_path) as img:
                info = img.info or {}
                png_text = {}
                for key, value in info.items():
                    if key == "exif":
                        continue
                    png_text[str(key)] = ImageUtils._coerce_metadata_value(value)
                if png_text:
                    metadata["png"] = png_text

                try:
                    exif_data = img.getexif()
                    if exif_data:
                        exif_map = {}
                        for tag_id, value in exif_data.items():
                            tag_name = ExifTags.TAGS.get(tag_id, str(tag_id))
                            exif_map[str(tag_name)] = ImageUtils._coerce_metadata_value(
                                value
                            )
                        if exif_map:
                            metadata["exif"] = exif_map
                except Exception:
                    pass
        except Exception as exc:
            logger.warning("Failed to extract embedded metadata: %s", exc)
        return metadata

    @staticmethod
    def resolve_picture_path(
        image_root: Optional[str], file_path: Optional[str]
    ) -> Optional[str]:
        """
        Resolve a stored picture path to an absolute file path.

        If file_path is already absolute it is returned unchanged.
        If file_path is relative it is joined with image_root.
        """
        if not file_path:
            return None
        if os.path.isabs(file_path):
            return file_path
        if not image_root:
            return file_path
        return os.path.join(image_root, file_path)

    @staticmethod
    def get_thumbnail_path(
        image_root: Optional[str], file_path: Optional[str]
    ) -> Optional[str]:
        """Return the expected thumbnail file path for a given picture path."""
        resolved = ImageUtils.resolve_picture_path(image_root, file_path)
        if not resolved:
            return None
        base, _ = os.path.splitext(resolved)
        return f"{base}_thumb{THUMBNAIL_EXTENSION}"

    @staticmethod
    def read_thumbnail_bytes(
        image_root: Optional[str], file_path: Optional[str]
    ) -> Optional[bytes]:
        """Read thumbnail bytes from disk, or return None if not found."""
        thumb_path = ImageUtils.get_thumbnail_path(image_root, file_path)
        if not thumb_path or not os.path.exists(thumb_path):
            return None
        try:
            with open(thumb_path, "rb") as handle:
                return handle.read()
        except Exception as exc:
            logger.warning("Failed to read thumbnail %s: %s", thumb_path, exc)
            return None

    @staticmethod
    def write_thumbnail_bytes(
        image_root: Optional[str], file_path: Optional[str], thumbnail: bytes
    ) -> Optional[str]:
        """Write thumbnail bytes to disk and return the path, or None on failure."""
        if not thumbnail:
            return None
        thumb_path = ImageUtils.get_thumbnail_path(image_root, file_path)
        if not thumb_path:
            return None
        try:
            os.makedirs(os.path.dirname(thumb_path), exist_ok=True)
            with open(thumb_path, "wb") as handle:
                handle.write(thumbnail)
            return thumb_path
        except Exception as exc:
            logger.warning("Failed to write thumbnail %s: %s", thumb_path, exc)
            return None

    @staticmethod
    def load_image_or_video_bgr(file_path: str) -> Optional[np.ndarray]:
        """Load an image or the first video frame as a BGR numpy array."""
        if not file_path or not os.path.exists(file_path):
            return None
        if VideoUtils.is_video_file(file_path):
            return VideoUtils._read_first_video_frame_bgr(file_path)

        img = cv2.imread(file_path)
        if img is not None:
            return img
        try:
            with Image.open(file_path) as pil_img:
                if pil_img.mode not in ("RGB", "L"):
                    pil_img = pil_img.convert("RGB")
                rgb = np.array(pil_img)
                return cv2.cvtColor(rgb, cv2.COLOR_RGB2BGR)
        except Exception:
            return None

    @staticmethod
    def _encode_thumbnail(pil_img: Image.Image) -> Optional[bytes]:
        """Encode a PIL image as a WebP thumbnail and return the bytes."""
        buf = BytesIO()
        try:
            pil_img.save(
                buf,
                format=THUMBNAIL_FORMAT,
                quality=THUMBNAIL_QUALITY,
                method=THUMBNAIL_WEBP_METHOD,
            )
        except Exception as exc:
            logger.error("Error encoding thumbnail bytes: %s", exc)
            return None
        return buf.getvalue()

    @staticmethod
    def clamp_bbox(bbox, width, height):
        """
        Clamp a bounding box [x_min, y_min, x_max, y_max] to image bounds.

        Returns a new list [x_min, y_min, x_max, y_max] or None if invalid.
        """
        if not bbox or len(bbox) != 4:
            return None
        x_min, y_min, x_max, y_max = [int(round(v)) for v in bbox]
        x_min = max(0, min(x_min, width - 1))
        y_min = max(0, min(y_min, height - 1))
        x_max = max(x_min + 1, min(x_max, width))
        y_max = max(y_min + 1, min(y_max, height))
        if x_max <= x_min or y_max <= y_min:
            return None
        return [x_min, y_min, x_max, y_max]

    @staticmethod
    def pad_image_to_square(pil_img: Image.Image, fill=0) -> Optional[Image.Image]:
        """Pad a PIL image to a square canvas while preserving content."""
        if pil_img is None:
            return None
        width, height = pil_img.size
        if width <= 0 or height <= 0:
            return None

        target = max(width, height)
        pad_x = max(0, target - width)
        pad_y = max(0, target - height)
        left = pad_x // 2
        right = pad_x - left
        top = pad_y // 2
        bottom = pad_y - top
        return ImageOps.expand(
            pil_img,
            border=(left, top, right, bottom),
            fill=fill,
        )

    @staticmethod
    def extract_created_at_from_metadata(
        image_bytes: bytes, fallback_file_path: str = None
    ) -> Optional[datetime]:
        """
        Try to extract the creation datetime from EXIF (for images), or from file
        metadata (for videos / filesystem).

        Returns a timezone-aware datetime in UTC, or None if not found.
        """
        try:
            with Image.open(BytesIO(image_bytes)) as img:
                exif_data = img.info.get("exif")
                if exif_data and piexif:
                    exif_dict = piexif.load(exif_data)
                    date_str = None
                    for tag in ("DateTimeOriginal", "DateTime", "DateTimeDigitised"):
                        val = exif_dict["0th"].get(
                            piexif.ImageIFD.__dict__.get(tag)
                        ) or exif_dict["Exif"].get(piexif.ExifIFD.__dict__.get(tag))
                        if val:
                            date_str = val.decode() if isinstance(val, bytes) else val
                            break
                    if date_str:

                        def _read_exif_offset() -> Optional[str]:
                            try:
                                exif_ifd = exif_dict.get("Exif") or {}
                                offset_tag_ids = [
                                    getattr(piexif.ExifIFD, "OffsetTimeOriginal", None),
                                    getattr(piexif.ExifIFD, "OffsetTime", None),
                                    getattr(
                                        piexif.ExifIFD, "OffsetTimeDigitized", None
                                    ),
                                    36880,
                                    36881,
                                    36882,
                                ]
                                for tag_id in offset_tag_ids:
                                    if tag_id is None:
                                        continue
                                    raw_val = exif_ifd.get(tag_id)
                                    if not raw_val:
                                        continue
                                    text = (
                                        raw_val.decode(errors="replace")
                                        if isinstance(raw_val, bytes)
                                        else str(raw_val)
                                    ).strip()
                                    if text:
                                        return text
                            except Exception:
                                return None
                            return None

                        try:
                            offset_text = _read_exif_offset()
                            if offset_text:
                                normalized_offset = offset_text.replace(" ", "")
                                dt = datetime.strptime(
                                    f"{date_str} {normalized_offset}",
                                    "%Y:%m:%d %H:%M:%S %z",
                                )
                                return dt.astimezone(timezone.utc)

                            dt = datetime.strptime(date_str, "%Y:%m:%d %H:%M:%S")
                            local_tz = (
                                datetime.now().astimezone().tzinfo or timezone.utc
                            )
                            return dt.replace(tzinfo=local_tz).astimezone(timezone.utc)
                        except Exception:
                            pass
        except Exception:
            pass

        if fallback_file_path and os.path.exists(fallback_file_path):
            try:
                ts = os.path.getmtime(fallback_file_path)
                dt = datetime.fromtimestamp(ts, tz=timezone.utc)
                return dt
            except Exception:
                pass
        return None

    @staticmethod
    def load_metadata(file_path):
        """
        Efficiently return ``(height, width, channels)`` for an image or video without
        loading full pixel data.
        """
        try:
            with Image.open(file_path) as img:
                w, h = img.size
                mode = img.mode
                if mode == "RGB":
                    c = 3
                elif mode == "L":
                    c = 1
                else:
                    c = len(img.getbands())
                return (h, w, c)
        except Exception:
            pass
        try:
            cap = cv2.VideoCapture(file_path)
            ret, frame = cap.read()
            cap.release()
            if ret and frame is not None:
                h, w = frame.shape[:2]
                c = frame.shape[2] if len(frame.shape) > 2 else 1
                return (h, w, c)
        except Exception:
            pass
        logger.error(f"Failed to read metadata for {file_path}")
        return None

    @staticmethod
    def load_image_or_video(file_path):
        """Load an image or video (first frame) and return an RGB numpy array."""
        try:
            try:
                with Image.open(file_path) as img:
                    return np.array(img.convert("RGB"))
            except Exception:
                pass
            frame = VideoUtils._read_first_video_frame_bgr(file_path)
            if frame is not None:
                frame_rgb = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                return frame_rgb
            else:
                raise ValueError("Could not read image or first frame from video.")
        except Exception as e:
            logger.error(f"Failed to load image at {file_path} for quality worker: {e}")
            return None

    @staticmethod
    def generate_thumbnail_bytes(img, size=(384, 384)) -> Optional[bytes]:
        """
        Crop to square (bottom-cropped for tall images) and resize longest edge.

        Accepts either a PIL Image or a numpy array (OpenCV BGR image).
        """
        try:
            if isinstance(img, Image.Image):
                pil_img = img.copy()
            else:
                pil_img = Image.fromarray(cv2.cvtColor(img, cv2.COLOR_BGR2RGB))
            if pil_img.mode not in ("RGB", "L"):
                pil_img = pil_img.convert("RGB")
            if pil_img.width != pil_img.height:
                side = min(pil_img.width, pil_img.height)
                if pil_img.height > pil_img.width:
                    left = 0
                    top = 0
                else:
                    left = int(round((pil_img.width - side) / 2.0))
                    top = 0
                right = left + side
                bottom = top + side
                pil_img = pil_img.crop((left, top, right, bottom))
            max_edge = max(pil_img.width, pil_img.height)
            if max_edge > size[0]:
                scale = size[0] / max_edge
                new_w = int(round(pil_img.width * scale))
                new_h = int(round(pil_img.height * scale))
                pil_img = pil_img.resize((new_w, new_h), resample=Image.LANCZOS)
            return ImageUtils._encode_thumbnail(pil_img)
        except Exception as e:
            logger.error(f"Error generating thumbnail bytes: {e}")
            return None

    @staticmethod
    def _calculate_sha256_digest(
        file_size: int,
        read_chunk,
        source_label: Optional[str] = None,
    ) -> str:
        """Compute a SHA-256 digest by either reading the whole file or sampling it."""
        chunk_size = 8192
        sample_count = 8
        whole_file_threshold = 128 * 1024

        sha256 = hashlib.sha256()
        if file_size <= whole_file_threshold:
            for offset in range(0, file_size, chunk_size):
                chunk = read_chunk(offset, chunk_size)
                if chunk:
                    sha256.update(chunk)
            digest = sha256.hexdigest()
            if source_label:
                logger.debug(f"WHOLE: {source_label} size={file_size} hash={digest}")
            else:
                logger.debug(f"WHOLE: size={file_size} hash={digest}")
            return digest

        offsets = [
            int(i * (file_size - chunk_size) / (sample_count - 1))
            for i in range(sample_count)
        ]
        for offset in offsets:
            chunk = read_chunk(offset, chunk_size)
            if chunk:
                sha256.update(chunk)
        digest = sha256.hexdigest()
        if source_label:
            logger.debug(f"SAMPLED: {source_label} size={file_size} hash={digest}")
        else:
            logger.debug(f"SAMPLED: hash={digest}")
        return digest

    @staticmethod
    def calculate_hash_from_file_path(file_path: str) -> str:
        """Compute a content hash for a file on disk."""
        file_size = os.path.getsize(file_path)
        with open(file_path, "rb") as f:

            def _read_chunk(offset, size):
                f.seek(offset)
                return f.read(size)

            return ImageUtils._calculate_sha256_digest(
                file_size=file_size,
                read_chunk=_read_chunk,
                source_label=file_path,
            )

    @staticmethod
    def calculate_hash_from_bytes(image_bytes: bytes) -> str:
        """Compute a content hash for raw image bytes."""
        file_size = len(image_bytes)

        def _read_chunk(offset, size):
            return image_bytes[offset : offset + size]

        return ImageUtils._calculate_sha256_digest(
            file_size=file_size,
            read_chunk=_read_chunk,
        )

    @staticmethod
    def create_picture_from_file(
        image_root_path: str,
        source_file_path: str,
        picture_uuid: Optional[str] = None,
        pixel_sha: Optional[str] = None,
    ) -> Picture:
        """
        Create a Picture from a file path, using metadata for created_at if available.
        """
        if not os.path.exists(source_file_path):
            raise ValueError(f"Source file path does not exist: {source_file_path}")
        with open(source_file_path, "rb") as f:
            image_bytes = f.read()
        created_at = ImageUtils.extract_created_at_from_metadata(
            image_bytes, fallback_file_path=source_file_path
        )
        return ImageUtils.create_picture_from_bytes(
            image_root_path=image_root_path,
            image_bytes=image_bytes,
            picture_uuid=picture_uuid,
            pixel_sha=pixel_sha,
            created_at=created_at,
        )

    @staticmethod
    def create_picture_from_bytes(
        image_root_path: str,
        image_bytes: bytes,
        picture_uuid: Optional[str] = None,
        pixel_sha: Optional[str] = None,
        created_at: Optional[str] = None,
    ) -> Picture:
        """Create a Picture from raw bytes, deriving metadata and saving the file."""
        if not pixel_sha:
            pixel_sha = ImageUtils.calculate_hash_from_bytes(image_bytes)

        inferred_ext = ""
        if picture_uuid:
            inferred_ext = os.path.splitext(str(picture_uuid))[1].lower().lstrip(".")

        img_format = None
        width = height = None
        thumbnail_bytes = None
        is_video = False
        try:
            with Image.open(BytesIO(image_bytes)) as img:
                img_format = img.format or "PNG"
                width, height = img.size
                thumbnail_bytes = ImageUtils.generate_thumbnail_bytes(img)
        except Exception:
            is_video = True

        if not is_video and (
            thumbnail_bytes is None or width is None or height is None
        ):
            raise ValueError("Failed to generate thumbnail for image bytes")

        if is_video:
            with tempfile.NamedTemporaryFile(delete=False, suffix=".mp4") as tmp:
                tmp.write(image_bytes)
                tmp_path = tmp.name
            cap = cv2.VideoCapture(tmp_path)
            ret, frame = cap.read()
            if not ret:
                logger.error("Could not read first frame from video for thumbnail.")
                raise ValueError("Failed to read first frame from video")
            else:
                height, width = frame.shape[:2]
                thumbnail_bytes = ImageUtils.generate_thumbnail_bytes(frame)
                if thumbnail_bytes is None:
                    raise ValueError("Failed to generate thumbnail for video")
            cap.release()
            if inferred_ext:
                img_format = inferred_ext.upper()
            else:
                img_format = "MP4"
            os.remove(tmp_path)

        if not picture_uuid:
            picture_uuid = str(uuid.uuid4()) + f".{img_format.lower()}"

        file_name = os.path.basename(picture_uuid)
        full_path = os.path.join(image_root_path, file_name)
        if os.path.exists(full_path):
            size_bytes = os.path.getsize(full_path)
        else:
            os.makedirs(image_root_path, exist_ok=True)
            with open(full_path, "wb") as f:
                f.write(image_bytes)
            size_bytes = len(image_bytes)

        if thumbnail_bytes:
            saved_thumb = ImageUtils.write_thumbnail_bytes(
                image_root_path, file_name, thumbnail_bytes
            )
            if not saved_thumb:
                logger.warning("Failed to persist thumbnail for %s", file_name)

        if not created_at:
            created_at = ImageUtils.extract_created_at_from_metadata(
                image_bytes, fallback_file_path=full_path
            )
        if not created_at:
            created_at = datetime.now(timezone.utc)

        pic = Picture(
            file_path=file_name,
            format=img_format,
            width=width,
            height=height,
            size_bytes=size_bytes,
            created_at=created_at,
            pixel_sha=pixel_sha,
        )
        return pic

    @staticmethod
    def cosine_similarity(a: bytes, b: bytes) -> float:
        """Compute cosine similarity between two embedding byte-strings."""
        try:
            if a is None or b is None:
                return 0.0
            arr_a = (
                np.frombuffer(a, dtype=np.float32)
                if isinstance(a, bytes)
                else np.array(a, dtype=np.float32)
            )
            arr_b = (
                np.frombuffer(b, dtype=np.float32)
                if isinstance(b, bytes)
                else np.array(b, dtype=np.float32)
            )
            if arr_a.shape != arr_b.shape or arr_a.size == 0:
                return 0.0
            dot = np.dot(arr_a, arr_b)
            norm_a = np.linalg.norm(arr_a)
            norm_b = np.linalg.norm(arr_b)
            if norm_a == 0 or norm_b == 0:
                return 0.0
            return float(dot / (norm_a * norm_b))
        except Exception as e:
            logger.warning(f"cosine_similarity error: {e}")
            return 0.0

    @classmethod
    def cosine_similarity_batch(cls, arr_a_list, arr_b_list):
        """
        Compute cosine similarity for two lists of np.ndarray feature vectors in batch.

        Returns a 1-D np.ndarray of similarities scaled to [0, 1].
        """
        arr_a = np.stack(arr_a_list)
        arr_b = np.stack(arr_b_list)
        arr_a_norm = arr_a / np.linalg.norm(arr_a, axis=1, keepdims=True)
        arr_b_norm = arr_b / np.linalg.norm(arr_b, axis=1, keepdims=True)
        sims = np.sum(arr_a_norm * arr_b_norm, axis=1)
        sims = 0.5 * (sims + 1.0)
        return sims



from __future__ import annotations

import os
import shutil
import socket
import tempfile
from importlib import metadata
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any

try:
    import onnxruntime as ort
except Exception:
    ort = None

try:
    import torch
except Exception:
    torch = None

from pixlvault.picture_tagger import PictureTagger


class StartupCheckError(Exception):
    def __init__(self, failures: list[str]):
        self.failures = list(failures)
        super().__init__(self._format_message())

    def _format_message(self) -> str:
        return "Startup checks failed:\n- " + "\n- ".join(self.failures)


@dataclass
class StartupCheckOutcome:
    hard_failures: list[str] = field(default_factory=list)
    warnings: list[str] = field(default_factory=list)
    notes: list[str] = field(default_factory=list)
    forced_cpu: bool = False


class StartupChecks:
    """Run startup preflight checks for server safety and readiness.

    Args:
        server_config (dict): Mutable server configuration dictionary.
        server_config_path (str): Path to server configuration file.
        logger: Logger instance used to report check results.
    """

    MIN_FREE_DISK_GB_DEFAULT = 1.0
    MIN_FREE_VRAM_MB_DEFAULT = 1024.0

    def __init__(self, server_config: dict, server_config_path: str, logger):
        self._server_config = server_config
        self._server_config_path = server_config_path
        self._logger = logger

    def run(self) -> dict[str, Any]:
        outcome = StartupCheckOutcome()

        self._check_config_sanity(outcome)
        self._check_image_root(outcome)
        self._check_database_path(outcome)
        self._check_free_disk_space(outcome)
        self._check_port_bindable(outcome)
        self._check_migration_assets(outcome)
        self._check_watch_folders(outcome)
        self._check_optional_dependencies(outcome)
        self._check_device_and_vram(outcome)

        for note in outcome.notes:
            self._logger.debug("[startup-check] %s", note)
        for warning in outcome.warnings:
            self._logger.warning("[startup-check] %s", warning)

        if outcome.hard_failures:
            self._logger.error(
                "[startup-check] Failed with %d hard failure(s).",
                len(outcome.hard_failures),
            )
            raise StartupCheckError(outcome.hard_failures)

        self._logger.info(
            "[startup-check] Passed (%d warning(s), forced_cpu=%s)",
            len(outcome.warnings),
            outcome.forced_cpu,
        )
        return {
            "warnings": list(outcome.warnings),
            "notes": list(outcome.notes),
            "forced_cpu": outcome.forced_cpu,
        }

    def _check_config_sanity(self, outcome: StartupCheckOutcome) -> None:
        host = self._server_config.get("host")
        if not isinstance(host, str) or not host.strip():
            outcome.hard_failures.append("Invalid server host in config.")

        port = self._server_config.get("port")
        try:
            port_int = int(port)
            if port_int < 0 or port_int > 65535:
                raise ValueError()
            self._server_config["port"] = port_int
        except Exception:
            outcome.hard_failures.append("Port must be an integer between 1 and 65535.")

        default_device = str(self._server_config.get("default_device", "cpu")).lower()
        if default_device not in {"cpu", "cuda", "gpu", "auto"}:
            outcome.hard_failures.append(
                "default_device must be one of: cpu, cuda, gpu, auto."
            )

        samesite = str(self._server_config.get("cookie_samesite", "Lax"))
        if samesite not in {"Lax", "Strict", "None"}:
            outcome.hard_failures.append(
                "cookie_samesite must be one of: Lax, Strict, None."
            )

        if ort is None:
            outcome.hard_failures.append(
                "onnxruntime is required but could not be imported."
            )

    def _check_image_root(self, outcome: StartupCheckOutcome) -> None:
        image_root = str(self._server_config.get("image_root") or "").strip()
        if not image_root:
            outcome.hard_failures.append("image_root is missing from config.")
            return

        try:
            os.makedirs(image_root, exist_ok=True)
        except Exception as exc:
            outcome.hard_failures.append(
                f"Unable to create image_root directory '{image_root}': {exc}"
            )
            return

        self._assert_dir_writable(image_root, "image_root", outcome)

    def _check_database_path(self, outcome: StartupCheckOutcome) -> None:
        image_root = str(self._server_config.get("image_root") or "").strip()
        if not image_root:
            return

        db_path = os.path.join(image_root, "vault.db")
        parent = os.path.dirname(db_path)
        self._assert_dir_writable(parent, "database directory", outcome)

        try:
            if os.path.exists(db_path):
                with open(db_path, "ab"):
                    pass
            else:
                with open(db_path, "ab"):
                    pass
                os.remove(db_path)
        except Exception as exc:
            outcome.hard_failures.append(
                f"Database file '{db_path}' is not writable: {exc}"
            )

    def _check_free_disk_space(self, outcome: StartupCheckOutcome) -> None:
        image_root = str(self._server_config.get("image_root") or "").strip()
        if not image_root:
            return

        min_free_gb = float(
            self._server_config.get("min_free_disk_gb", self.MIN_FREE_DISK_GB_DEFAULT)
        )
        self._server_config["min_free_disk_gb"] = min_free_gb

        try:
            usage = shutil.disk_usage(image_root)
            free_gb = usage.free / float(1024**3)
        except Exception as exc:
            outcome.hard_failures.append(
                f"Unable to determine free disk space for '{image_root}': {exc}"
            )
            return

        if free_gb < min_free_gb:
            outcome.hard_failures.append(
                f"Insufficient disk space at '{image_root}': {free_gb:.2f} GB free, requires >= {min_free_gb:.2f} GB."
            )
        else:
            outcome.notes.append(
                f"Disk space OK at image_root: {free_gb:.2f} GB free (threshold {min_free_gb:.2f} GB)."
            )

    def _check_port_bindable(self, outcome: StartupCheckOutcome) -> None:
        host = str(self._server_config.get("host", "localhost"))
        port = int(self._server_config.get("port", 8000))

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            sock.bind((host, port))
        except Exception as exc:
            outcome.hard_failures.append(f"Server cannot bind to {host}:{port}: {exc}")
        finally:
            sock.close()

    def _check_migration_assets(self, outcome: StartupCheckOutcome) -> None:
        module_dir = Path(__file__).resolve().parent
        repo_root = module_dir.parent

        candidate_locations = [
            (repo_root / "alembic.ini", repo_root / "migrations"),
            (module_dir / "alembic.ini", module_dir / "migrations"),
        ]

        for candidate_ini, candidate_migrations in candidate_locations:
            if candidate_ini.exists() and candidate_migrations.exists():
                outcome.notes.append(
                    f"Alembic assets found at {candidate_ini} and {candidate_migrations}."
                )
                return

        expected = " or ".join(
            f"({candidate_ini}, {candidate_migrations})"
            for candidate_ini, candidate_migrations in candidate_locations
        )
        outcome.hard_failures.append(f"Alembic assets not found. Expected {expected}.")

    def _check_watch_folders(self, outcome: StartupCheckOutcome) -> None:
        watch_folders = self._server_config.get("watch_folders") or []
        if not isinstance(watch_folders, list):
            outcome.hard_failures.append("watch_folders must be a list.")
            return

        for entry in watch_folders:
            folder = entry
            if isinstance(entry, dict):
                folder = entry.get("folder")
            if not folder:
                continue
            if not os.path.exists(folder):
                outcome.warnings.append(f"Watch folder does not exist: {folder}")

    def _check_optional_dependencies(self, outcome: StartupCheckOutcome) -> None:
        openssl_path = shutil.which("openssl")
        if self._server_config.get("require_ssl", False) and not openssl_path:
            outcome.hard_failures.append(
                "require_ssl is enabled but openssl is not available on PATH."
            )
        elif not openssl_path:
            outcome.warnings.append("Optional dependency missing: openssl")

        if not shutil.which("nvidia-smi"):
            outcome.warnings.append(
                "Optional GPU utility missing: nvidia-smi (GPU telemetry may be reduced)."
            )

    def _has_onnxruntime_conflict(self) -> bool:
        """Return True if both onnxruntime and onnxruntime-gpu are installed.

        Having both packages installed causes ONNX to silently use CPU inference
        even when a CUDA-capable GPU is present.
        """
        cpu_installed = False
        gpu_installed = False
        for package_name in ("onnxruntime", "onnxruntime-gpu"):
            try:
                metadata.distribution(package_name)
                if package_name == "onnxruntime":
                    cpu_installed = True
                else:
                    gpu_installed = True
            except metadata.PackageNotFoundError:
                continue
            except Exception:
                continue
        return cpu_installed and gpu_installed

    def _check_device_and_vram(self, outcome: StartupCheckOutcome) -> None:
        device_value = str(self._server_config.get("default_device", "cpu")).lower()
        if device_value == "gpu":
            device_value = "cuda"

        is_auto_mode = device_value == "auto"
        is_explicit_gpu = device_value == "cuda"

        if device_value == "cpu":
            PictureTagger.FORCE_CPU = True
            outcome.notes.append(
                "default_device is set to cpu in config; using CPU inference."
            )
            return

        if torch is None:
            self._handle_gpu_check_failure(
                outcome,
                is_auto_mode,
                is_explicit_gpu,
                "PyTorch is unavailable; forcing CPU inference.",
                "PyTorch is unavailable while default_device is set to cuda.",
            )
            return

        if not torch.cuda.is_available():
            self._handle_gpu_check_failure(
                outcome,
                is_auto_mode,
                is_explicit_gpu,
                "CUDA is unavailable; forcing CPU inference.",
                "CUDA is unavailable while default_device is set to cuda.",
            )
            return

        providers = []
        try:
            providers = ort.get_available_providers() if ort is not None else []
        except Exception:
            providers = []
        if "CUDAExecutionProvider" not in providers:
            provider_list = ", ".join(providers) if providers else "none"
            onnx_package = self._detect_onnxruntime_package()
            gpu_arch_note = self._detect_gpu_arch_note()
            remediation = self._onnx_cuda_remediation_hint(onnx_package, gpu_arch_note)
            conflict_note = (
                " Both 'onnxruntime' and 'onnxruntime-gpu' are installed — "
                "this conflict likely caused the fallback to CPU. "
                "Fix with: pip uninstall -y onnxruntime && pip install onnxruntime-gpu."
                if self._has_onnxruntime_conflict()
                else ""
            )
            # If PyTorch CUDA is available the GPU works — only the ONNX WD14 tagger
            # will fall back to CPU.  Treat this as a warning, not a hard failure.
            if is_explicit_gpu:
                outcome.warnings.append(
                    "ONNX CUDAExecutionProvider unavailable while default_device is set to cuda "
                    f"(available providers: {provider_list}; package: {onnx_package}){gpu_arch_note}. "
                    "PyTorch CUDA will still be used for all non-ONNX inference; "
                    f"the WD14 tagger ONNX model will run on CPU.{conflict_note} "
                    f"{remediation}"
                )
            else:
                outcome.warnings.append(
                    "ONNX CUDAExecutionProvider unavailable "
                    f"(available providers: {provider_list}; package: {onnx_package}){gpu_arch_note}. "
                    "PyTorch CUDA will still be used for all non-ONNX inference; "
                    f"the WD14 tagger ONNX model will run on CPU.{conflict_note} {remediation}"
                )
            # Don't return — continue with the VRAM check so other GPU paths work.

        min_free_vram_mb = float(
            self._server_config.get(
                "min_free_vram_mb",
                self.MIN_FREE_VRAM_MB_DEFAULT,
            )
        )
        self._server_config["min_free_vram_mb"] = min_free_vram_mb

        try:
            free_bytes, total_bytes = torch.cuda.mem_get_info()
            free_mb = free_bytes / float(1024**2)
            total_mb = total_bytes / float(1024**2)
        except Exception as exc:
            self._handle_gpu_check_failure(
                outcome,
                is_auto_mode,
                is_explicit_gpu,
                f"Unable to read VRAM availability ({exc}); forcing CPU inference.",
                f"Unable to read VRAM availability while default_device is set to cuda: {exc}",
            )
            return

        if free_mb < min_free_vram_mb:
            self._handle_gpu_check_failure(
                outcome,
                is_auto_mode,
                is_explicit_gpu,
                (
                    f"Insufficient free VRAM ({free_mb:.0f} MB of {total_mb:.0f} MB; "
                    f"requires >= {min_free_vram_mb:.0f} MB); forcing CPU inference."
                ),
                (
                    f"Insufficient free VRAM while default_device is set to cuda "
                    f"({free_mb:.0f} MB of {total_mb:.0f} MB; requires >= {min_free_vram_mb:.0f} MB)."
                ),
            )
            return

        PictureTagger.FORCE_CPU = False
        if device_value == "auto":
            self._server_config["default_device"] = "auto"
        outcome.notes.append(
            f"GPU check passed ({free_mb:.0f} MB free VRAM); using CUDA inference."
        )

    def _force_cpu_with_warning(
        self,
        outcome: StartupCheckOutcome,
        warning: str,
        is_auto_mode: bool = False,
    ) -> None:
        PictureTagger.FORCE_CPU = True
        # Only persist "cpu" when it was an explicit user choice.  When in auto
        # mode we keep "auto" in the config so that the next startup re-evaluates
        # CUDA availability (e.g. after upgrading the CUDA runtime or drivers).
        if not is_auto_mode:
            self._server_config["default_device"] = "cpu"
        outcome.forced_cpu = True
        outcome.warnings.append(warning)

    def _handle_gpu_check_failure(
        self,
        outcome: StartupCheckOutcome,
        is_auto_mode: bool,
        is_explicit_gpu: bool,
        fallback_warning: str,
        explicit_gpu_failure: str,
    ) -> None:
        if is_explicit_gpu and not is_auto_mode:
            outcome.hard_failures.append(explicit_gpu_failure)
            return
        self._force_cpu_with_warning(
            outcome, fallback_warning, is_auto_mode=is_auto_mode
        )

    def _detect_gpu_arch_note(self) -> str:
        """Return a human-readable note if the GPU arch may be unsupported by ORT."""
        if torch is None or not torch.cuda.is_available():
            return ""
        try:
            major, minor = torch.cuda.get_device_capability(0)
            sm = major * 10 + minor
            # ORT 1.x releases lag behind new GPU architectures.
            # sm >= 120 = Blackwell (RTX 5xxx) — not included in ORT until a later release.
            if sm >= 120:
                name = torch.cuda.get_device_name(0)
                return (
                    f" — GPU {name} (sm_{major}{minor}, Blackwell) may not be supported "
                    "by the installed onnxruntime-gpu; upgrade to a newer ORT release or "
                    "build from source"
                )
        except Exception:
            # Best-effort diagnostic only: if we cannot query GPU capability,
            # fall back to no additional note rather than failing startup.
            return ""
        return ""

    def _detect_onnxruntime_package(self) -> str:
        for package_name in ("onnxruntime-gpu", "onnxruntime"):
            try:
                distribution = metadata.distribution(package_name)
                return f"{package_name} {distribution.version}"
            except metadata.PackageNotFoundError:
                continue
            except Exception:
                continue
        return "unknown"

    def _onnx_cuda_remediation_hint(
        self, onnx_package: str, gpu_arch_note: str = ""
    ) -> str:
        config_hint = (
            f"Server config path: {self._server_config_path}.\n"
            "Set `default_device` to `cpu` or `auto` there to avoid strict CUDA startup checks."
        )
        if onnx_package.startswith("onnxruntime "):
            return (
                "Detected CPU-only ONNX Runtime.\nInstall GPU support with "
                "`pip uninstall -y onnxruntime && pip install onnxruntime-gpu`.\n"
                f"{config_hint}"
            )
        if onnx_package == "unknown":
            return (
                "Verify ONNX Runtime installation and ensure CUDA provider support is installed.\n"
                f"(for pip: `pip install onnxruntime-gpu`). {config_hint}"
            )
        return (
            "Ensure ONNX Runtime CUDA dependencies are installed and accessible on this machine.\n"
            f"{config_hint}"
        )

    def _assert_dir_writable(
        self,
        dir_path: str,
        label: str,
        outcome: StartupCheckOutcome,
    ) -> None:
        try:
            os.makedirs(dir_path, exist_ok=True)
            with tempfile.NamedTemporaryFile(
                mode="wb",
                dir=dir_path,
                delete=False,
                prefix="pixlvault-startup-check-",
            ) as handle:
                handle.write(b"ok")
                temp_path = handle.name
            os.remove(temp_path)
        except Exception as exc:
            outcome.hard_failures.append(f"{label} '{dir_path}' is not writable: {exc}")



import cv2
from PIL import Image
import numpy as np
import os
import torch

from pixlvault.pixl_logging import get_logger

IMAGE_SIZE = 448

logger = get_logger(__name__)


class ImageLoadingDatasetPrepper(torch.utils.data.Dataset):
    def __init__(self, image_paths):
        self.images = image_paths

    def __len__(self):
        return len(self.images)

    def __getitem__(self, idx):
        img_path = str(self.images[idx])
        ext = os.path.splitext(img_path)[1].lower()
        if ext in [".jpg", ".jpeg", ".png", ".webp", ".bmp", ".heic", ".heif"]:
            try:
                image = Image.open(img_path).convert("RGB")
                image = self._preprocess_image(image)
            except Exception as e:
                logger.error(f"Could not load image path: {img_path}, error: {e}")
                return None
            return (image, img_path)
        elif ext in [".mp4", ".avi", ".mov", ".mkv"]:
            # Extract only the first frame from video and treat it as one image
            try:
                import cv2

                cap = cv2.VideoCapture(img_path)
                if not cap.isOpened():
                    logger.error(f"Could not open video file: {img_path}")
                    return None
                frame_count = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
                if frame_count < 1:
                    logger.error(f"No frames found in video: {img_path}")
                    cap.release()
                    return None
                frame_indices = [0]
                images = []
                for idx_frame in frame_indices:
                    cap.set(cv2.CAP_PROP_POS_FRAMES, idx_frame)
                    ret, frame = cap.read()
                    if not ret:
                        logger.error(
                            f"Could not read frame {idx_frame} from video: {img_path}"
                        )
                        continue
                    frame_rgb = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                    pil_img = Image.fromarray(frame_rgb)
                    prepped = self._preprocess_image(pil_img)
                    images.append((prepped, f"{img_path}#frame{idx_frame}"))
                cap.release()
                if not images:
                    logger.error(f"No frames extracted from video: {img_path}")
                    return None
                # Return a single (image, img_path#frameX) tuple list for compatibility
                return images
            except Exception as e:
                logger.error(f"Could not process video file: {img_path}, error: {e}")
                return None
        else:
            logger.error(f"Unsupported file extension for tagging: {img_path}")
            return None

    @staticmethod
    def _preprocess_image(image, image_size=IMAGE_SIZE):
        image = np.array(image)
        image = image[:, :, ::-1]  # RGB->BGR

        # pad to square
        size = max(image.shape[0:2])
        pad_x = size - image.shape[1]
        pad_y = size - image.shape[0]
        pad_l = pad_x // 2
        pad_t = pad_y // 2
        image = np.pad(
            image,
            ((pad_t, pad_y - pad_t), (pad_l, pad_x - pad_l), (0, 0)),
            mode="constant",
            constant_values=255,
        )

        image = cv2.resize(
            image, (image_size, image_size), interpolation=cv2.INTER_AREA
        )

        image = image.astype(np.float32)
        return image



"""Smart score calculation utilities."""

from typing import List, Optional

import numpy as np

from pixlvault.pixl_logging import get_logger

logger = get_logger(__name__)


class SmartScoreUtils:
    """Utility methods for computing smart scores over picture candidates."""

    @staticmethod
    def calculate_smart_score_batch_numpy(
        candidates: List[dict],
        good_anchors: List[dict],
        bad_anchors: List[dict],
        config: Optional[dict] = None,
    ) -> np.ndarray:
        """
        Calculate smart scores for a batch of candidates using numpy.

        Args:
            candidates: List of dicts with 'id', 'embedding' (numpy array), 'aesthetic_score'.
            good_anchors: List of dicts with 'embedding', 'score'.
            bad_anchors: List of dicts with 'embedding', 'score'.
            config: Config dict overrides.

        Returns:
            np.ndarray: Array of floating point scores corresponding to candidates.
        """
        cfg = {
            "w_good": 0.13,
            "w_bad": 0.07,
            "w_aest": 0.35,
            "w_sharpness": 0.35,
            "w_resolution": 0.10,
            "w_penalised_tag": 0.40,
            "penalised_tag_cap": 3.5,
            "topk": 3,
            "sim_knee": 0.3,
            "sim_power": 1.5,
            "bad_sim_knee": 0.3,
            "bad_sim_power": 1.5,
            "aest_min": 2.0,
            "aest_max": 7.0,
            "res_min_mpx": 0.2,
            "res_max_mpx": 4.0,
            "res_use_log": True,
            "batch_normalize": True,
            "batch_normalize_lo_pct": 5.0,
            "batch_normalize_hi_pct": 95.0,
        }
        if config:
            cfg.update(config)

        if not candidates:
            return np.array([])

        target_shape = None
        for c in candidates:
            emb = c.get("embedding")
            if isinstance(emb, np.ndarray) and emb.ndim == 1 and emb.size > 0:
                target_shape = emb.shape
                break

        if target_shape is None:
            return np.zeros(len(candidates))

        invalid_embeddings = 0
        cand_vecs = []
        for c in candidates:
            emb = c.get("embedding")
            if (
                isinstance(emb, np.ndarray)
                and emb.ndim == 1
                and emb.shape == target_shape
            ):
                cand_vecs.append(emb.astype(np.float32, copy=False))
            else:
                cand_vecs.append(np.zeros(target_shape, dtype=np.float32))
                invalid_embeddings += 1

        if invalid_embeddings:
            logger.warning(
                "[SMART SCORE] %s candidates had invalid/mismatched embeddings; using zeros.",
                invalid_embeddings,
            )

        raw_aest = np.array(
            [float(c.get("aesthetic_score") or 5.0) for c in candidates]
        )

        a_min = cfg.get("aest_min", 3.0)
        a_max = cfg.get("aest_max", 7.0)
        denom = max(0.1, a_max - a_min)
        cand_aest = np.clip((raw_aest - a_min) / denom, 0.0, 1.0)

        M_cand = np.stack(cand_vecs)
        scores = np.zeros(len(candidates))

        def norm_weight(s):
            effective = s if s > 0 else 2.5
            return max(0, min(1, (effective - 1) / 4.0))

        def sim01_batch(A, B):
            return 0.5 * (1 + np.dot(A, B.T))

        # Good anchors
        good_component = np.zeros(len(candidates))
        M_good = None
        if good_anchors:
            good_pairs = [
                (a["embedding"], a.get("score", 0))
                for a in good_anchors
                if isinstance(a.get("embedding"), np.ndarray)
                and a["embedding"].ndim == 1
                and a["embedding"].shape == target_shape
            ]
            if good_pairs:
                good_vecs = [p[0] for p in good_pairs]
                good_weights = np.array([norm_weight(p[1]) for p in good_pairs])
                M_good = np.stack(good_vecs)

        if good_anchors and M_good is not None:
            sims = sim01_batch(M_cand, M_good)
            knee = float(cfg["sim_knee"])
            power = float(cfg["sim_power"])
            smooth = np.clip((sims - knee) / max(1e-6, 1.0 - knee), 0.0, 1.0) ** power
            weighted = smooth * good_weights
            K = min(cfg["topk"], weighted.shape[1])
            if K > 0:
                if K < weighted.shape[1]:
                    topk = -np.partition(-weighted, K - 1, axis=1)[:, :K]
                else:
                    topk = weighted
                avg_good = np.mean(topk, axis=1)
                good_component = cfg["w_good"] * avg_good
                scores += good_component

        # Bad anchors
        bad_component = np.zeros(len(candidates))
        M_bad = None
        if bad_anchors:
            bad_pairs = [
                (a["embedding"], a.get("score", 0))
                for a in bad_anchors
                if isinstance(a.get("embedding"), np.ndarray)
                and a["embedding"].ndim == 1
                and a["embedding"].shape == target_shape
            ]
            if bad_pairs:
                bad_vecs = [p[0] for p in bad_pairs]
                bad_weights = np.array([1.0 - norm_weight(p[1]) for p in bad_pairs])
                M_bad = np.stack(bad_vecs)

        if bad_anchors and M_bad is not None:
            sims = sim01_batch(M_cand, M_bad)
            bad_knee = float(cfg["bad_sim_knee"])
            bad_power = float(cfg["bad_sim_power"])
            smooth_bad = (
                np.clip((sims - bad_knee) / max(1e-6, 1.0 - bad_knee), 0.0, 1.0)
                ** bad_power
            )
            weighted = smooth_bad * bad_weights
            K = min(cfg["topk"], weighted.shape[1])
            if K > 0:
                if K < weighted.shape[1]:
                    topk = -np.partition(-weighted, K - 1, axis=1)[:, :K]
                else:
                    topk = weighted
                avg_bad = np.mean(topk, axis=1)
                bad_component = cfg["w_bad"] * avg_bad
                scores -= bad_component

        # Aesthetic
        aest_component = cfg["w_aest"] * cand_aest
        scores += aest_component

        # Resolution (megapixels)
        widths = np.array([c.get("width") or 0 for c in candidates], dtype=np.float32)
        heights = np.array([c.get("height") or 0 for c in candidates], dtype=np.float32)
        mpx = (widths * heights) / 1_000_000.0
        res_min = float(cfg.get("res_min_mpx", 0.2) or 0.2)
        res_max = float(cfg.get("res_max_mpx", 4.0) or 4.0)
        res_min = max(0.01, res_min)
        res_max = max(res_min + 0.01, res_max)
        if cfg.get("res_use_log", True):
            res_vals = np.log10(np.clip(mpx, 1e-6, None))
            res_min_val = np.log10(res_min)
            res_max_val = np.log10(res_max)
        else:
            res_vals = mpx
            res_min_val = res_min
            res_max_val = res_max
        denom_res = max(0.001, res_max_val - res_min_val)
        res_norm = np.clip((res_vals - res_min_val) / denom_res, 0.0, 1.0)
        res_component = cfg["w_resolution"] * res_norm
        scores += res_component

        # Subject sharpness (high if at least one region is in focus)
        sharpness_vals = np.array(
            [c.get("sharpness") for c in candidates], dtype=np.float32
        )
        sharpness_vals = np.where(np.isfinite(sharpness_vals), sharpness_vals, np.nan)
        sharpness_vals = np.where(sharpness_vals < 0, np.nan, sharpness_vals)
        sharpness_vals = np.where(np.isnan(sharpness_vals), 0.5, sharpness_vals)
        sharpness_vals = np.clip(sharpness_vals, 0.0, 1.0)
        sharpness_component = cfg["w_sharpness"] * sharpness_vals
        scores += sharpness_component

        # Penalised tags
        penalised_counts = np.array(
            [float(c.get("penalised_tag_count") or 0) for c in candidates]
        )
        penalised_equivalent = np.clip(
            penalised_counts / 5.0, 0.0, float(cfg["penalised_tag_cap"])
        )
        penalised_component = cfg["w_penalised_tag"] * penalised_equivalent
        scores -= penalised_component

        # Soft batch normalization: stretch scores so the distribution fills [0,1]
        # using robust percentiles so individual outliers don't dominate.
        if cfg.get("batch_normalize", True) and len(scores) >= 4:
            lo_pct = float(cfg.get("batch_normalize_lo_pct", 5.0))
            hi_pct = float(cfg.get("batch_normalize_hi_pct", 95.0))
            p_lo = np.percentile(scores, lo_pct)
            p_hi = np.percentile(scores, hi_pct)
            span = p_hi - p_lo
            if span > 0.01:
                scores = (scores - p_lo) / span

        # Rescale [0, 1] to [1, 5]
        clipped = np.clip(scores, 0.0, 1.0)
        final_scores = 1.0 + (clipped * 4.0)

        try:
            for idx, candidate in enumerate(candidates):
                penalised_count = float(penalised_counts[idx])
                final_score = float(final_scores[idx])
                if penalised_count <= 0 and final_score <= 1.5:
                    logger.debug(
                        "[SMART SCORE][MIN] id=%s raw=%.4f clipped=%.4f final=%.4f "
                        "good=%.4f bad=%.4f aest=%.4f sharpness=%.4f res=%.4f "
                        "penalised=%.4f mpx=%.4f w=%s h=%s",
                        candidate.get("id"),
                        float(scores[idx]),
                        float(clipped[idx]),
                        final_score,
                        float(good_component[idx]),
                        float(bad_component[idx]),
                        float(aest_component[idx]),
                        float(sharpness_component[idx]),
                        float(res_component[idx]),
                        float(penalised_component[idx]),
                        float(mpx[idx]),
                        candidate.get("width"),
                        candidate.get("height"),
                    )
        except Exception as exc:
            logger.warning("[SMART SCORE][MIN] logging failed: %s", exc)

        return final_scores



import gc
import pytest
import random
import time
import tempfile
import os
from fastapi.testclient import TestClient
from pixlvault.server import Server

BACKEND_URL = "http://localhost:9537"


@pytest.mark.parametrize(
    "params",
    [
        {},
        {"sort": "SCORE", "descending": True},
        {"sort": "DATE", "descending": False},
    ],
)
def test_order_stability(params):
    """
    For each set of parameters, repeatedly query the backend and check that the returned image IDs are always in the same order.
    """
    with tempfile.TemporaryDirectory() as temp_dir:
        image_root = os.path.join(temp_dir, "images")
        os.makedirs(image_root, exist_ok=True)
        server_config_path = os.path.join(temp_dir, "server-config.json")

        with Server(server_config_path) as server:
            server.vault.import_default_data(True)
            client = TestClient(server.api)

            resp = client.post(
                "/login", json={"username": "testuser", "password": "testpassword"}
            )
            assert resp.status_code == 200
            print("Login Response: ", resp.json())  # Should include "session_id"
            assert "session_id" in client.cookies, (
                "session_id cookie not set after login"
            )

            resp = client.get("/protected")
            print("Protected Response: ", resp.json())

            first_ids = []

            for i in range(0, 3):
                time.sleep(random.uniform(0.01, 0.05))
                resp = client.get("/pictures", params={**params})
                assert resp.status_code == 200, (
                    f"Backend returned {resp.status_code} for params {params}"
                )
                data = resp.json()
                ids = [img["id"] for img in data if "id" in img]
                if i == 0:
                    first_ids = ids
                else:
                    assert ids == first_ids, (
                        f"Order not stable for params {params}: {ids} != {first_ids}"
                    )
    gc.collect()



import base64
import sys
import numpy as np

from datetime import datetime

from enum import Enum, auto, IntEnum
from sqlalchemy import String, desc, func
from sqlalchemy.orm import load_only, selectinload
from sqlalchemy.types import LargeBinary
from sqlmodel import (
    Column,
    DateTime,
    SQLModel,
    Field,
    Relationship,
    exists,
    select,
    Session,
)
from typing import ClassVar, Optional, List, TYPE_CHECKING

from .face import Face
from .picture_set import PictureSet, PictureSetMember
from .picture_stack import PictureStack
from .quality import Quality
from .tag import Tag

from pixlvault.pixl_logging import get_logger

if TYPE_CHECKING:
    from .character import Character
    from .picture_likeness import PictureLikeness


# Configure logging for the module
logger = get_logger(__name__)


# Class for sorting mechanisms (replaces Enum)
class SortMechanism:
    class Keys(Enum):
        DATE = auto()
        SCORE = auto()
        CHARACTER_LIKENESS = auto()
        PICTURE_STACKS = auto()
        IMAGE_SIZE = auto()
        SMART_SCORE = auto()

    MECHANISMS = {
        Keys.DATE: {
            "field": "created_at",
            "description": "Date Created",
        },
        Keys.SCORE: {
            "field": "score",
            "description": "Score",
        },
        Keys.SMART_SCORE: {
            "field": None,
            "description": "Smart Score",
        },
        Keys.CHARACTER_LIKENESS: {
            "field": "character_likeness",
            "description": "Similarity to",
        },
        Keys.PICTURE_STACKS: {
            "field": "id",
            "description": "Likeness Groups",
        },
        Keys.IMAGE_SIZE: {
            "field": None,  # Special case, not a direct field
            "description": "Image Size",
        },
    }

    def __init__(self, key, descending: bool = True):
        self.key = key
        self.descending = descending

    @property
    def field(self):
        return self.MECHANISMS[self.key]["field"]

    @classmethod
    def all(cls):
        mechanisms = []
        for key, data in cls.MECHANISMS.items():
            data = {"key": str(key.name), **data}
            mechanisms.append(data)
        return mechanisms

    @classmethod
    def from_string(cls, key_string: str, descending: bool = True) -> "SortMechanism":
        # Try by name
        if key_string in cls.Keys.__members__:
            return SortMechanism(cls.Keys[key_string], descending=descending)

        raise ValueError(f"{key_string!r} is not a valid SortMechanism")


class LikenessParameter(IntEnum):
    SIZE_BIN = 0
    BRIGHTNESS = 1
    CONTRAST = 2
    EDGE_DENSITY = 3
    NOISE_LEVEL = 4
    ASPECT_RATIO = 5
    PHASH_PREFIX = 6
    DATE = 7
    COLORFULNESS = 8
    LUMINANCE_ENTROPY = 9
    DOMINANT_HUE = 10


LIKENESS_PARAMETER_SENTINEL = -1.0


def _default_likeness_parameters() -> np.ndarray:
    return np.full(
        len(LikenessParameter), LIKENESS_PARAMETER_SENTINEL, dtype=np.float32
    )


class ExportType(Enum):
    FULL = "full"
    FACE = "face"

    @classmethod
    def from_string(cls, value: str) -> "ExportType":
        d = (value or "").lower()
        for member in cls:
            if member.value == d:
                return member
        return cls.FULL


class Picture(SQLModel, table=True):
    ExportType: ClassVar[type["ExportType"]] = ExportType
    id: int = Field(default=None, primary_key=True)
    file_path: Optional[str] = None
    description: Optional[str] = None
    format: Optional[str] = None
    width: Optional[int] = None
    height: Optional[int] = None
    size_bytes: Optional[int] = None
    size_bin_index: Optional[int] = Field(default=None, index=True)
    created_at: Optional[datetime] = Field(
        default=None, sa_column=Column("created_at", type_=DateTime, nullable=True)
    )
    imported_at: Optional[datetime] = Field(
        default=None,
        sa_column=Column("imported_at", type_=DateTime, nullable=True, index=True),
    )
    text_embedding: Optional[np.ndarray] = Field(
        sa_column=Column("text_embedding", LargeBinary, default=None, nullable=True)
    )
    image_embedding: Optional[np.ndarray] = Field(
        sa_column=Column("image_embedding", LargeBinary, default=None, nullable=True)
    )
    likeness_parameters: Optional[np.ndarray] = Field(
        default_factory=_default_likeness_parameters,
        sa_column=Column(
            "likeness_parameters", LargeBinary, default=None, nullable=True
        ),
    )
    perceptual_hash: Optional[str] = Field(
        default=None,
        sa_column=Column("perceptual_hash", String, default=None, nullable=True),
    )
    thumbnail_left: Optional[int] = Field(default=None)
    thumbnail_top: Optional[int] = Field(default=None)
    thumbnail_side: Optional[int] = Field(default=None)
    score: Optional[int] = None
    aesthetic_score: Optional[float] = None
    pixel_sha: Optional[str] = Field(default=None, index=True)
    deleted: bool = Field(default=False, index=True)
    stack_id: Optional[int] = Field(
        default=None, foreign_key="picturestack.id", index=True
    )
    stack_position: Optional[int] = Field(default=None, index=True)

    # Relationships
    quality: Optional["Quality"] = Relationship(
        back_populates="picture",
        sa_relationship_kwargs={
            # face Quality rows share the same picture_id as their parent picture,
            # so we must filter to face_id IS NULL to get only picture-level quality.
            "primaryjoin": "(Quality.picture_id == Picture.id) & (Quality.face_id == None)",
            "foreign_keys": "[Quality.picture_id]",
        },
    )
    faces: List["Face"] = Relationship(
        back_populates="picture",
        sa_relationship_kwargs={
            "overlaps": "characters",
            "cascade": "all, delete-orphan",
            "passive_deletes": True,
        },
    )
    tags: List["Tag"] = Relationship(
        back_populates="picture",
        sa_relationship_kwargs={
            "cascade": "all, delete-orphan",
            "passive_deletes": True,
            "foreign_keys": "[Tag.picture_id]",
        },
    )
    characters: List["Character"] = Relationship(  # Many-to-many via Face
        back_populates="pictures",
        link_model=Face,
        sa_relationship_kwargs={"overlaps": "faces,picture,character"},
    )
    picture_sets: List["PictureSet"] = Relationship(
        back_populates="members", link_model=PictureSetMember
    )
    stack: Optional["PictureStack"] = Relationship(back_populates="pictures")

    likeness_a: List["PictureLikeness"] = Relationship(
        back_populates="picture_a",
        sa_relationship_kwargs={
            "primaryjoin": "Picture.id==PictureLikeness.picture_id_a",
            "cascade": "all, delete-orphan",
            "passive_deletes": True,
        },
    )
    likeness_b: List["PictureLikeness"] = Relationship(
        back_populates="picture_b",
        sa_relationship_kwargs={
            "primaryjoin": "Picture.id==PictureLikeness.picture_id_b",
            "cascade": "all, delete-orphan",
            "passive_deletes": True,
        },
    )

    class Config:
        arbitrary_types_allowed = True

    def __hash__(self):
        # Use the unique id for hashing
        return hash(self.id)

    def __eq__(self, other):
        # Compare by id for equality
        if isinstance(other, Picture):
            return self.id == other.id
        return False

    def text_embedding_data(self):
        """
        Returns a structured dict for embedding: description, tags, and character info.
        """
        data = {
            "description": self.description or None,
            "tags": [
                tag.tag
                for tag in getattr(self, "tags", [])
                if getattr(tag, "tag", None)
            ],
            "characters": [],
        }
        for character in getattr(self, "characters", []):
            char_info = {
                "name": getattr(character, "name", None),
                "description": getattr(character, "description", None),
            }
            data["characters"].append(char_info)
        return data

    @classmethod
    def semantic_search(
        cls: "Picture",
        session: Session,
        query: str,
        query_words: List[str],
        text_to_embedding: callable,
        clip_text_to_embedding: callable = None,
        fuzzy_weight: float = 0.5,
        embedding_weight: float = 0.5,
        threshold: float = 0.0,
        offset: int = 0,
        limit: int = sys.maxsize,
        format: Optional[List[str]] = None,
        select_fields: Optional[List[str]] = None,
        include_deleted: bool = False,
        only_deleted: bool = False,
        include_unimported: bool = True,
        candidate_ids: Optional[List[int]] = None,
    ) -> List["Picture"]:
        """
        Hybrid semantic search: combines fuzzy tag search (levenshtein SQL function) and embedding similarity (cosine_similarity SQL function).
        Orders by combined score in SQL.
        """
        if candidate_ids is not None and not candidate_ids:
            return []
        # 1. Generate SBERT embedding for tag search (Text-to-Text)
        query_embedding = text_to_embedding(query)
        if query_embedding is None:
            logger.warning("Semantic search: Failed to generate SBERT embedding.")
            query_embedding_bytes = None
        else:
            query_embedding_bytes = query_embedding.tobytes()

        # 2. Generate CLIP embedding for visual search (Text-to-Image)
        if clip_text_to_embedding:
            clip_query_embedding = clip_text_to_embedding(query)
            clip_query_embedding_bytes = (
                clip_query_embedding.tobytes()
                if clip_query_embedding is not None
                else None
            )
        else:
            clip_query_embedding_bytes = None

        logger.debug(
            f"Performing semantic search for query='{query}' and query_words={query_words} with fuzzy_weight={fuzzy_weight}, embedding_weight={embedding_weight}"
        )

        query_str = " ".join(query_words)
        # Subquery: calculate levenshtein distance for all tags of each picture
        tag_query = select(
            Tag.picture_id,
            func.levenshtein_with_id(
                func.group_concat(Tag.tag, " "), query_str, Tag.picture_id
            ).label("min_tag_dist"),
        ).group_by(Tag.picture_id)

        if candidate_ids:
            tag_query = tag_query.where(Tag.picture_id.in_(candidate_ids))

        tag_subq = tag_query.subquery()

        # Calculate cosine similarity for both text (tags) and image (visuals) embeddings
        if query_embedding_bytes:
            text_sim = (
                func.coalesce(
                    func.cosine_similarity(cls.text_embedding, query_embedding_bytes),
                    0.0,
                )
                * 2.0
            )
        else:
            text_sim = 0.0

        if clip_query_embedding_bytes:
            # Boost logic: CLIP similarity for unrelated text-image pairs is low (0.1-0.2).
            # A good match is often 0.25-0.35.
            # Fuzzy match is 0.0 to 1.0 (usually 1.0 for matches).
            # To make CLIP comparable and impactful, we multiply by a factor (e.g., 2.5).
            # This brings 0.3 -> 0.75, which can rival a messy tag match.
            image_sim = (
                func.coalesce(
                    func.cosine_similarity(
                        cls.image_embedding, clip_query_embedding_bytes
                    ),
                    0.0,
                )
                * 2.5
            )
        else:
            image_sim = 0.0

        # Combined embedding score: average of text and image similarity to capture both explicit tags and visual concepts
        embedding_score_raw = (text_sim + image_sim) / 2.0
        embedding_score = func.sqrt(func.max(0.0, embedding_score_raw))

        raw_fuzzy_score = func.max(
            0.0, 1.0 - func.coalesce(tag_subq.c.min_tag_dist, 1.0)
        )
        fuzzy_score = func.pow(raw_fuzzy_score, 1.5)

        # Main query: join pictures with tag_subq, compute combined score
        stmt = (
            select(
                cls,
                (fuzzy_weight * fuzzy_score + embedding_weight * embedding_score).label(
                    "combined_score"
                ),
                fuzzy_score.label("fuzzy_score"),
                embedding_score.label("embedding_score"),
                tag_subq.c.min_tag_dist.label(
                    "min_tag_dist"
                ),  # Explicitly include min_tag_dist
            )
            .outerjoin(tag_subq, cls.id == tag_subq.c.picture_id)
            .order_by(desc("combined_score"))
            .offset(offset)
            .limit(limit)
        )

        if only_deleted:
            stmt = stmt.where(cls.deleted.is_(True))
        elif not include_deleted:
            stmt = stmt.where(cls.deleted.is_(False))

        if not include_unimported:
            stmt = stmt.where(cls.imported_at.is_not(None))

        # Apply select_fields logic (like in find)
        if select_fields:
            select_fields = list(set(select_fields) | {"id"})

            # Use load_only for scalar fields
            scalar_attrs = [
                getattr(cls, field)
                for field in cls.scalar_fields().intersection(select_fields)
            ]
            if scalar_attrs:
                stmt = stmt.options(load_only(*scalar_attrs))
            # Use selectinload for relationships present in select_fields
            rel_attrs = [
                getattr(cls, field)
                for field in cls.relationship_fields().intersection(select_fields)
            ]
            for rel_attr in rel_attrs:
                stmt = stmt.options(selectinload(rel_attr))

        if format:
            stmt = stmt.where(cls.format.in_(format))

        if candidate_ids:
            stmt = stmt.where(cls.id.in_(candidate_ids))

        results = session.exec(stmt).all()

        if results:
            top_rows = results[:20]
            header = f"{'rank':>4} {'id':>6} {'combined':>9} {'fuzzy':>7} {'embed':>7} {'min_tag':>8}"
            lines = [header]
            for idx, row in enumerate(top_rows, start=1):
                pic = row[0]
                combined_score = row[1] if row[1] is not None else 0.0
                fuzzy_val = row[2] if row[2] is not None else 0.0
                embed_val = row[3] if row[3] is not None else 0.0
                min_tag_val = row[4] if row[4] is not None else 0.0
                pic_id = getattr(pic, "id", None)
                lines.append(
                    f"{idx:>4} {pic_id:>6} {float(combined_score):>9.4f} {float(fuzzy_val):>7.4f} {float(embed_val):>7.4f} {float(min_tag_val):>8.4f}"
                )
            logger.info(
                "Semantic search score breakdown (top %d) query=%r:\n%s",
                len(top_rows),
                query,
                "\n".join(lines),
            )

        # Log tag contribution for each result
        for result in results:
            logger.debug(
                f"Picture ID: {result.Picture.id}, "
                f"Tag contribution: min_tag_dist={getattr(result, 'min_tag_dist', 'N/A')}, "
                f"fuzzy_score={getattr(result, 'fuzzy_score', 'N/A')}, "
                f"embedding_score={getattr(result, 'embedding_score', 'N/A')}, "
                f"combined_score={getattr(result, 'combined_score', 'N/A')}"
            )
        output = []
        for row in results:
            pic, combined_score, _, _ = (
                row[0],
                row[1],
                row[2],
                row[3],
            )
            if combined_score and combined_score >= threshold:
                output.append((pic, combined_score))
        return output

    @staticmethod
    def serialize_with_likeness(picture_and_score):
        pic, score = picture_and_score
        d = pic.to_serializable_dict()
        d["likeness_score"] = score
        return d

    @classmethod
    def find(
        cls,
        session,
        *,
        sort_mech: Optional[SortMechanism] = None,
        offset: int = 0,
        limit: int = sys.maxsize,
        select_fields: Optional[List[str]] = None,
        format: Optional[List[str]] = None,
        include_deleted: bool = False,
        only_deleted: bool = False,
        include_unimported: bool = True,
        **search,
    ) -> List["Picture"]:
        """
        Find pictures based on provided filters.
        """
        query = select(cls)

        logger.debug("Got search parameters: %s", search)
        if select_fields:
            # Always include 'id' in select_fields
            select_fields = list(set(select_fields) | {"id"})
            # Use load_only for scalar fields
            scalar_attrs = [
                getattr(cls, field)
                for field in cls.scalar_fields().intersection(select_fields)
            ]
            if scalar_attrs:
                query = query.options(load_only(*scalar_attrs))
            # Use selectinload for relationships present in select_fields
            rel_attrs = [
                getattr(cls, field)
                for field in cls.relationship_fields().intersection(select_fields)
            ]
            for rel_attr in rel_attrs:
                query = query.options(selectinload(rel_attr))

        if only_deleted:
            query = query.where(cls.deleted.is_(True))
        elif not include_deleted:
            query = query.where(cls.deleted.is_(False))

        if not include_unimported:
            query = query.where(cls.imported_at.is_not(None))

        for attr, value in search.items():
            if hasattr(cls, attr):
                if isinstance(value, list):
                    query = query.where(getattr(cls, attr).in_(value))
                else:
                    query = query.where(getattr(cls, attr) == value)

        if format:
            query = query.where(cls.format.in_(format))

        if sort_mech:
            if sort_mech.key == SortMechanism.Keys.IMAGE_SIZE:
                # Sort by width * height
                if sort_mech.descending:
                    query = query.order_by(
                        (cls.width * cls.height).desc(), cls.id.desc()
                    )
                else:
                    query = query.order_by((cls.width * cls.height).asc(), cls.id.asc())
            else:
                field_name = sort_mech.field
                field = getattr(cls, field_name, None)
                if field is not None:
                    if sort_mech.descending:
                        query = query.order_by(field.desc(), cls.id.desc())
                    else:
                        query = query.order_by(field.asc(), cls.id.asc())
        if offset > 0 or limit != sys.maxsize:
            query = query.offset(offset).limit(limit)

        results = session.exec(query).all()
        return results

    @classmethod
    def metadata_fields(cls):
        """
        Return a list of simple scalar fields
        """
        return cls.scalar_fields() - cls.large_binary_fields()

    @classmethod
    def grid_fields(cls):
        """
        Return a minimal set of fields for grid listing.
        """
        return {
            "id",
            "width",
            "height",
            "format",
            "score",
            "created_at",
            "stack_id",
            "stack_position",
        }

    @classmethod
    def scalar_fields(cls):
        """
        Return a list of simple scalar fields
        """
        return set(cls.__table__.columns.keys())

    @classmethod
    def relationship_fields(cls):
        """
        Return a list of relationship fields
        """
        return set(Picture.__mapper__.relationships.keys())

    @classmethod
    def large_binary_fields(cls):
        """
        Return a list of LargeBinary fields
        """
        return {
            field.name
            for field in cls.__table__.columns
            if isinstance(field.type, LargeBinary)
        }

    def to_serializable_dict(self):
        """
        Returns a dict suitable for JSON serialization, encoding all large binary fields as base64 if present.
        """
        d = self.model_dump()
        for field in self.large_binary_fields():
            val = d.get(field, None)
            if val is not None:
                try:
                    if isinstance(val, np.ndarray):
                        val_bytes = val.tobytes()
                    elif isinstance(val, (bytes, bytearray)):
                        val_bytes = val
                    else:
                        val_bytes = bytes(val)
                    d[field] = base64.b64encode(val_bytes).decode("utf-8")
                except Exception:
                    d[field] = None
        return d

    @classmethod
    def clear_field(cls, session, picture_ids, field_name: str):
        pictures = cls.find(session=session, id=picture_ids, select_fields=[field_name])
        for pic in pictures:
            if hasattr(pic, field_name):
                setattr(pic, field_name, None)
        session.add_all(pictures)
        session.commit()

    @classmethod
    def find_unassigned(
        cls,
        session,
        sort_mech: Optional[SortMechanism] = None,
        offset: int = 0,
        limit: int = sys.maxsize,
        format: list[str] | None = None,
        metadata_fields: list[str] | None = None,
    ):
        query = select(Picture)
        unassigned_condition = ~exists(
            select(Face.id).where(
                Face.picture_id == Picture.id,
                Face.character_id.is_not(None),
            )
        )
        not_in_set_condition = ~exists(
            select(PictureSetMember.picture_id).where(
                PictureSetMember.picture_id == Picture.id
            )
        )
        query = query.where(
            unassigned_condition,
            not_in_set_condition,
            Picture.deleted.is_(False),
        )

        if format:
            query = query.where(Picture.format.in_(format))

        select_fields = metadata_fields or cls.metadata_fields()
        if select_fields:
            select_fields = list(set(select_fields) | {"id"})
            scalar_attrs = [
                getattr(Picture, field)
                for field in Picture.scalar_fields().intersection(select_fields)
            ]
            if scalar_attrs:
                query = query.options(load_only(*scalar_attrs))
            rel_attrs = [
                getattr(Picture, field)
                for field in Picture.relationship_fields().intersection(select_fields)
            ]
            for rel_attr in rel_attrs:
                query = query.options(selectinload(rel_attr))

        if sort_mech:
            if sort_mech.key == SortMechanism.Keys.IMAGE_SIZE:
                order_expr = Picture.width * Picture.height
                query = query.order_by(
                    order_expr.desc() if sort_mech.descending else order_expr.asc(),
                    Picture.id.desc() if sort_mech.descending else Picture.id.asc(),
                )
            else:
                field = getattr(Picture, sort_mech.field, None)
                if field is not None:
                    query = query.order_by(
                        field.desc() if sort_mech.descending else field.asc(),
                        Picture.id.desc() if sort_mech.descending else Picture.id.asc(),
                    )

        if offset > 0 or limit != sys.maxsize:
            query = query.offset(offset).limit(limit)

        return session.exec(query).all()

    @staticmethod
    def fetch_features(session, picture_ids):
        """
        Fetch faces for a list of picture IDs.
        Returns:
            faces_by_pic: dict[picture_id, list[Face]]
            hands_by_pic: dict (empty, hand detection removed)
            face_tags_by_face: dict (empty, face tags removed)
            hand_tags_by_hand: dict (empty, hand tags removed)
        """
        faces = session.exec(select(Face).where(Face.picture_id.in_(picture_ids))).all()

        faces_by_pic = {}
        for face in faces:
            faces_by_pic.setdefault(face.picture_id, []).append(face)

        return (
            faces_by_pic,
            {},
            {},
            {},
        )



import base64
import json
import subprocess
from datetime import date, datetime
from sqlmodel import SQLModel

# Add import for SQLAlchemy CollectionAdapter
try:
    from sqlalchemy.orm.collections import CollectionAdapter
except ImportError:
    CollectionAdapter = None


def safe_model_dict(obj) -> dict:
    """
    Recursively create a safe, serializable dict from any SQLModel instance, dict, or SQLAlchemy adapter.
    - Encodes bytes fields as base64.
    - Parses JSON/text fields ending with '_'.
    - Recurses into SQLModel relationships, lists, dicts, and adapters.
    """
    if CollectionAdapter and isinstance(obj, CollectionAdapter):
        # Convert SQLAlchemy adapter to list
        return [safe_model_dict(v) for v in list(obj)]
    if isinstance(obj, dict):
        result = {}
        for k, v in obj.items():
            result[k] = safe_model_dict(v)
        return result
    if isinstance(obj, list):
        return [safe_model_dict(v) for v in obj]
    if isinstance(obj, (int, float, str, bool)) or obj is None:
        return obj
    if isinstance(obj, (datetime, date)):
        return obj
    result = {}
    for field, value in obj.__dict__.items():
        if field.startswith("_sa"):
            continue
        if isinstance(value, bytes):
            result[field] = base64.b64encode(value).decode("utf-8")
        elif field.endswith("_") and isinstance(value, str):
            try:
                result[field[:-1]] = json.loads(value)
            except Exception:
                result[field[:-1]] = value
        elif CollectionAdapter and isinstance(value, CollectionAdapter):
            result[field] = [safe_model_dict(v) for v in list(value)]
        elif isinstance(value, SQLModel):
            result[field] = safe_model_dict(value)
        elif isinstance(value, list):
            result[field] = [safe_model_dict(v) for v in value]
        elif isinstance(value, dict):
            result[field] = safe_model_dict(value)
        else:
            result[field] = value
    return result


def default_max_vram_gb() -> float:
    """Return default VRAM budget in GB: min(4GB, 25% of available VRAM).

    Falls back to 4GB when VRAM cannot be detected.
    """
    try:
        output = subprocess.check_output(
            [
                "nvidia-smi",
                "--query-gpu=memory.total",
                "--format=csv,noheader,nounits",
            ],
            stderr=subprocess.DEVNULL,
            text=True,
        )
        totals_mb = []
        for line in output.splitlines():
            value = line.strip()
            if not value:
                continue
            totals_mb.append(int(float(value)))
        total_mb = sum(totals_mb)
        if total_mb <= 0:
            return 4.0
        quarter_gb = (total_mb / 1024.0) / 4.0
        return round(min(4.0, quarter_gb), 2)
    except Exception:
        return 4.0


def serialize_user_config(user) -> dict:
    from pixlvault.db_models import (
        User,
        DEFAULT_SMART_SCORE_PENALIZED_TAGS,
        DEFAULT_SMART_SCORE_PENALIZED_TAG_WEIGHT,
    )

    default_user = User()
    source = user or default_user

    allowed_fields = {
        "description",
        "sort",
        "descending",
        "columns",
        "sidebar_thumbnail_size",
        "show_stars",
        "show_face_bboxes",
        "show_hand_bboxes",
        "show_format",
        "show_resolution",
        "show_problem_icon",
        "date_format",
        "theme_mode",
        "comfyui_url",
        "similarity_character",
        "stack_strictness",
        "apply_tag_filter",
        "keep_models_in_memory",
        "max_vram_gb",
    }

    config = {
        key: (
            getattr(source, key)
            if getattr(source, key) is not None
            else getattr(default_user, key)
        )
        for key in allowed_fields
    }
    config["expand_all_stacks"] = (
        getattr(source, "show_stacks")
        if getattr(source, "show_stacks") is not None
        else getattr(default_user, "show_stacks")
    )

    allowed_sidebar_sizes = tuple(range(32, 65, 8))
    sidebar_size = _thumbnail_size(config.get("sidebar_thumbnail_size"))
    if sidebar_size is None:
        sidebar_size = default_user.sidebar_thumbnail_size
    if sidebar_size not in allowed_sidebar_sizes:
        sidebar_size = min(allowed_sidebar_sizes, key=lambda v: abs(v - sidebar_size))
    config["sidebar_thumbnail_size"] = sidebar_size

    config["smart_score_penalised_tags"] = _smart_score_penalised_tags(
        getattr(source, "smart_score_penalised_tags", None),
        DEFAULT_SMART_SCORE_PENALIZED_TAGS,
        default_weight=DEFAULT_SMART_SCORE_PENALIZED_TAG_WEIGHT,
    )
    config["hidden_tags"] = _normalize_hidden_tags(getattr(source, "hidden_tags", None))
    config["sort_order"] = config["sort"]
    if config.get("max_vram_gb") is None:
        config["max_vram_gb"] = default_max_vram_gb()
    return config


def apply_user_config_patch(user, patch_data) -> bool:
    from pixlvault.db_models import DEFAULT_SMART_SCORE_PENALIZED_TAG_WEIGHT

    allowed_fields = {
        "description",
        "sort",
        "descending",
        "columns",
        "sidebar_thumbnail_size",
        "show_stars",
        "show_face_bboxes",
        "show_hand_bboxes",
        "show_format",
        "show_resolution",
        "show_problem_icon",
        "expand_all_stacks",
        "show_stacks",
        "date_format",
        "theme_mode",
        "comfyui_url",
        "similarity_character",
        "stack_strictness",
        "smart_score_penalised_tags",
        "hidden_tags",
        "apply_tag_filter",
        "keep_models_in_memory",
        "max_vram_gb",
    }

    allowed_date_formats = {
        "locale",
        "iso",
        "eu",
        "us",
        "ymd-slash",
        "ymd-dot",
        "ymd-jp",
    }
    allowed_theme_modes = {"light", "dark"}

    updated = False
    for key, value in patch_data.items():
        if key not in allowed_fields:
            raise ValueError(f"Key '{key}' does not exist in config.")
        if key in {"expand_all_stacks", "show_stacks"}:
            new_value = bool(value)
            if user.show_stacks != new_value:
                user.show_stacks = new_value
                updated = True
            continue
        if key == "similarity_character":
            if value in ("", None, "null"):
                new_value = None
            elif isinstance(value, str) and value.isdigit():
                new_value = int(value)
            else:
                new_value = value
            if user.similarity_character != new_value:
                user.similarity_character = new_value
                updated = True
            continue
        if key == "comfyui_url":
            if value in ("", None, "null"):
                new_value = None
            else:
                new_value = str(value).strip()
            if user.comfyui_url != new_value:
                user.comfyui_url = new_value
                updated = True
            continue
        if key == "smart_score_penalised_tags":
            if value in ("", None):
                new_value = None
            else:
                d = _smart_score_penalised_tags(
                    value,
                    None,
                    allow_empty=True,
                    default_weight=DEFAULT_SMART_SCORE_PENALIZED_TAG_WEIGHT,
                )
                if d is None:
                    raise ValueError(
                        "smart_score_penalised_tags must be a JSON list or object"
                    )
                new_value = json.dumps(d)
            if user.smart_score_penalised_tags != new_value:
                user.smart_score_penalised_tags = new_value
                updated = True
            continue
        if key == "hidden_tags":
            if value in ("", None, "null"):
                normalized = []
            else:
                normalized = _normalize_hidden_tags(value)
                if normalized is None:
                    raise ValueError("hidden_tags must be a JSON list of strings")
            new_value = json.dumps(normalized)
            if user.hidden_tags != new_value:
                user.hidden_tags = new_value
                updated = True
            continue
        if key == "apply_tag_filter":
            if value in ("", None, "null"):
                new_value = False
            else:
                new_value = bool(value)
            if user.apply_tag_filter != new_value:
                user.apply_tag_filter = new_value
                updated = True
            continue
        if key == "keep_models_in_memory":
            if value in ("", None, "null"):
                new_value = True
            else:
                new_value = bool(value)
            if user.keep_models_in_memory != new_value:
                user.keep_models_in_memory = new_value
                updated = True
            continue
        if key == "max_vram_gb":
            if value in ("", None, "null"):
                new_value = None
            else:
                new_value = float(value)
                if new_value <= 0:
                    raise ValueError("max_vram_gb must be greater than 0")
            if user.max_vram_gb != new_value:
                user.max_vram_gb = new_value
                updated = True
            continue
        if key == "date_format":
            if value in ("", None, "null"):
                new_value = None
            else:
                new_value = str(value)
            if new_value is not None and new_value not in allowed_date_formats:
                raise ValueError("date_format is not a supported value")
            if user.date_format != new_value:
                user.date_format = new_value
                updated = True
            continue
        if key == "theme_mode":
            if value in ("", None, "null"):
                new_value = None
            else:
                new_value = str(value)
            if new_value is not None and new_value not in allowed_theme_modes:
                raise ValueError("theme_mode is not a supported value")
            if user.theme_mode != new_value:
                user.theme_mode = new_value
                updated = True
            continue
        if key == "stack_strictness":
            if value in ("", None, "null"):
                new_value = None
            else:
                new_value = float(value)
            if user.stack_strictness != new_value:
                user.stack_strictness = new_value
                updated = True
            continue
        if key == "columns":
            new_value = int(value)
            if user.columns != new_value:
                user.columns = new_value
                updated = True
            continue
        if key == "sidebar_thumbnail_size":
            new_value = _thumbnail_size(value)
            if new_value is None:
                continue
            allowed_sizes = tuple(range(32, 65, 8))
            if new_value not in allowed_sizes:
                new_value = min(allowed_sizes, key=lambda v: abs(v - new_value))
            if user.sidebar_thumbnail_size != new_value:
                user.sidebar_thumbnail_size = new_value
                updated = True
            continue
        current_value = getattr(user, key, None)
        if current_value != value:
            setattr(user, key, value)
            updated = True
    return updated


def serialize_tag_objects(tags: list | None, empty_sentinel: str = "") -> list[dict]:
    items = []
    for tag in tags or []:
        if not tag or getattr(tag, "tag", None) in (None, empty_sentinel):
            continue
        items.append({"id": getattr(tag, "id", None), "tag": tag.tag})
    return items


def _thumbnail_size(value):
    if value is None:
        return None
    if isinstance(value, str):
        if value.lower() == "default":
            return None
        if value.isdigit():
            return int(value)
        return None
    if isinstance(value, (int, float)):
        return int(value)
    return None


def _smart_score_penalised_tags(
    value,
    fallback=None,
    allow_empty: bool = False,
    default_weight: int = 3,
):
    if value is None:
        return fallback

    tags = None
    if isinstance(value, str):
        try:
            tags = json.loads(value)
        except Exception:
            return fallback
    else:
        tags = value

    if isinstance(tags, list):
        d = {}
        for tag in tags:
            if tag is None:
                continue
            clean = str(tag).strip().lower()
            if not clean:
                continue
            d[clean] = default_weight
    elif isinstance(tags, dict):
        d = {}
        for tag, weight in tags.items():
            if tag is None:
                continue
            clean = str(tag).strip().lower()
            if not clean:
                continue
            try:
                weight_value = int(float(weight))
            except (TypeError, ValueError):
                weight_value = default_weight
            weight_value = max(1, min(5, weight_value))
            existing = d.get(clean)
            if existing is None or weight_value > existing:
                d[clean] = weight_value
    else:
        return fallback

    if d:
        return d
    return {} if allow_empty else fallback


def _normalize_hidden_tags(value):
    if value is None:
        return []

    if isinstance(value, str):
        try:
            tags = json.loads(value)
        except Exception:
            return None
    else:
        tags = value

    if isinstance(tags, dict):
        tags = list(tags.keys())
    if not isinstance(tags, list):
        return None

    cleaned = []
    seen = set()
    for tag in tags:
        if tag is None:
            continue
        clean = str(tag).strip().lower()
        if not clean or clean in seen:
            continue
        seen.add(clean)
        cleaned.append(clean)
    return cleaned



import json
import math

from sqlalchemy.orm import joinedload
from sqlmodel import (
    Column,
    ForeignKey,
    Integer,
    select,
    String,
    SQLModel,
    Field,
    Relationship,
    UniqueConstraint,
)
from typing import List, Optional, TYPE_CHECKING

from .quality import Quality

if TYPE_CHECKING:
    from .picture import Picture
    from .character import Character


class Face(SQLModel, table=True):
    id: int = Field(default=None, primary_key=True)

    picture_id: int = Field(
        sa_column=Column(
            Integer,
            ForeignKey("picture.id", ondelete="CASCADE"),
            index=True,
        ),
        default=None,
    )
    frame_index: int = Field(default=0)
    face_index: int = Field(default=0)

    character_id: Optional[int] = Field(
        sa_column=Column(Integer, ForeignKey("character.id"), default=None, index=True)
    )
    bbox_: Optional[str] = Field(sa_column=Column("bbox", String, default=None))
    features: Optional[bytes] = None

    # Relationships
    quality: Optional[Quality] = Relationship(
        back_populates="face",
        sa_relationship_kwargs={
            "cascade": "all, delete-orphan",
            "passive_deletes": True,
        },
    )
    picture: Optional["Picture"] = Relationship(
        back_populates="faces", sa_relationship_kwargs={"overlaps": "character"}
    )
    character: Optional["Character"] = Relationship(
        back_populates="faces", sa_relationship_kwargs={"overlaps": "picture"}
    )

    __table_args__ = (UniqueConstraint("picture_id", "frame_index", "face_index"),)

    def __init__(self, *args, bbox=None, **kwargs):
        super().__init__(*args, **kwargs)
        if bbox is not None:
            self.bbox = bbox

    @property
    def bbox(self) -> Optional[List[int]]:
        """
        Return the bounding box as a list of integers, or None if not set.
        """
        if self.bbox_:
            return json.loads(self.bbox_)
        return None

    @bbox.setter
    def bbox(self, bbox: List[int]):
        """
        Set the bounding box from a list of integers.
        """
        self.bbox_ = json.dumps(bbox)

    @property
    def width(self) -> Optional[float]:
        """
        Return the width of the face bounding box, or 0.0 if bbox is not set.
        """
        if self.bbox and len(self.bbox) == 4:
            return self.bbox[2] - self.bbox[0]
        return 0.0

    @property
    def height(self) -> Optional[float]:
        """
        Return the height of the face bounding box, or 0.0 if bbox is not set.
        """
        if self.bbox and len(self.bbox) == 4:
            return self.bbox[3] - self.bbox[1]
        return 0.0

    @classmethod
    def find(cls, session, **filters) -> Optional["Face"]:
        """
        Find faces by picture_id, frame_index, and/or face_index.
        Supports passing a list for picture_id (uses IN_ if so).
        """
        query = select(cls).options(joinedload(cls.quality)).where(cls.face_index != -1)
        for attr, value in filters.items():
            if hasattr(cls, attr):
                col = getattr(cls, attr)
                if attr == "picture_id" and isinstance(value, list):
                    query = query.where(col.in_(value))
                else:
                    query = query.where(col == value)

        faces = session.exec(query).all()
        return faces

    @staticmethod
    def expand_face_bbox(
        bbox: List[int],
        picture_width: int,
        picture_height: int,
        expansion_fraction: float,
    ) -> List[int]:
        """
        Expand the bounding box by a given expansion fraction and align to 64-pixel boundaries.
        Args:
            bbox: List or tuple of [x_min, y_min, x_max, y_max]
            expansion_fraction: Fraction to expand the bbox on each side
        Returns:
            Expanded bbox as [x_min, y_min, x_max, y_max]
        """
        if bbox is None or len(bbox) != 4:
            return bbox
        x_min, y_min, x_max, y_max = bbox

        width = x_max - x_min
        height = y_max - y_min

        def round64(val):
            return int(math.ceil(val / 64.0) * 64)

        new_width = round64(width + width * expansion_fraction)
        new_height = round64(height + height * expansion_fraction)

        width_expansion = new_width - width
        height_expansion = new_height - height

        x_min = max(0, int(round(x_min - width_expansion / 2)))
        x_max = min(picture_width, int(round(x_min + new_width)))
        y_min = max(0, int(round(y_min - height_expansion / 2)))
        y_max = min(picture_height, int(round(y_min + new_height)))

        return [
            x_min,
            y_min,
            x_max,
            y_max,
        ]



import tempfile
import os
import time

from pixlvault.pixl_logging import get_logger
from pixlvault.server import Server
from pixlvault.tasks.task_type import TaskType
from pixlvault.db_models.picture import Picture
from pixlvault.db_models.picture_likeness import PictureLikeness, PictureLikenessQueue
from pixlvault.utils.image_processing.image_utils import ImageUtils
from sqlalchemy import func
from sqlmodel import select


# Configure logging for the module
logger = get_logger(__name__)


def test_likeness_worker():
    with tempfile.TemporaryDirectory() as temp_dir:
        image_root = os.path.join(temp_dir, "images")
        os.makedirs(image_root, exist_ok=True)
        server_config_path = os.path.join(temp_dir, "server-config.json")
        print("Launching server for likeness worker test...")
        with Server(server_config_path) as server:
            print("Server started for likeness worker test.")
            server.vault.import_default_data(add_tagger_test_images=True)
            # Get all pictures
            pictures = server.vault.db.run_task(
                lambda session: Picture.find(
                    session, select_fields=Picture.metadata_fields()
                )
            )
            assert pictures and len(pictures) >= 2, (
                "No pictures found in the database. Test requires at least two pictures."
            )

            quality_pictures = [pic for pic in pictures if pic.file_path]
            assert quality_pictures, (
                "Expected at least one picture with file_path for quality calculation."
            )
            quality_paths = [
                ImageUtils.resolve_picture_path(
                    server.vault.db.image_root, pic.file_path
                )
                for pic in quality_pictures
            ]
            assert any(os.path.exists(path) for path in quality_paths), (
                "Expected at least one quality-processable picture file to exist on disk."
            )

            quality_futures = []
            for pic in quality_pictures:
                future = server.vault.get_worker_future(
                    TaskType.QUALITY, Picture, pic.id, "quality"
                )
                quality_futures.append(future)
            logger.info(f"Queued {len(quality_futures)} quality computations.")
            # Start the quality worker
            # Wait for all quality computations to complete
            timeout = time.time() + 120
            for future in quality_futures:
                future.result(timeout=timeout - time.time())

            logger.info("All picture quality computations completed.")

            def fetch_missing_prereqs(session):
                rows = session.exec(
                    select(Picture.id).where(
                        (Picture.image_embedding.is_(None))
                        | (Picture.likeness_parameters.is_(None))
                        | (Picture.perceptual_hash.is_(None))
                    )
                ).all()
                return [row for row in rows]

            timeout = time.time() + 240
            missing = server.vault.db.run_task(fetch_missing_prereqs)
            while missing and time.time() < timeout:
                time.sleep(0.5)
                missing = server.vault.db.run_task(fetch_missing_prereqs)
            assert not missing, (
                "Timed out waiting for likeness prerequisites for picture ids: "
                f"{missing}"
            )
            # Get all unique pairs (a < b)
            pairs = []
            ids = sorted([pic.id for pic in pictures])
            for i, a in enumerate(ids):
                for b in ids[i + 1 :]:
                    pairs.append((a, b))
            logger.info(f"Queued {len(pairs)} likeness pairs for processing.")
            # Start the likeness worker

            def fetch_queue_remaining(session):
                return session.exec(
                    select(func.count()).select_from(PictureLikenessQueue)
                ).one()

            timeout = time.time() + 120
            remaining = server.vault.db.run_task(fetch_queue_remaining)
            while remaining and time.time() < timeout:
                time.sleep(0.5)
                remaining = server.vault.db.run_task(fetch_queue_remaining)
            assert not remaining, (
                f"Timed out waiting for likeness queue to drain. Remaining={remaining}"
            )
            # Check that all likeness results are present
            likeness_results = server.vault.db.run_task(
                lambda session: PictureLikeness.find(session)
            )
            result_pairs = set(
                (r.picture_id_a, r.picture_id_b) for r in likeness_results
            )
            if not likeness_results:
                logger.warning(
                    "No likeness results were produced; gating may have filtered all pairs."
                )
            for a, b in result_pairs:
                assert (a, b) in pairs, f"Unexpected likeness pair produced: ({a}, {b})"

            # Print table of likeness scores with descriptions
            pic_map = {pic.id: pic for pic in pictures}
            logger.info("\nLikeness Table:")
            logger.info(f"{'Desc A':<30} | {'Desc B':<30} | {'Likeness':<10}")
            logger.info("-" * 110)
            for r in likeness_results:
                pic_a = pic_map.get(r.picture_id_a)
                pic_b = pic_map.get(r.picture_id_b)
                desc_a = (pic_a.description or "") if pic_a else "?"
                desc_b = (pic_b.description or "") if pic_b else "?"
                logger.info(f"{desc_a:<30} | {desc_b:<30} | {r.likeness:<10.4f}")



import os
import tempfile
import gc
import sys

from PIL import Image
from io import BytesIO
from fastapi.testclient import TestClient

from pixlvault.server import Server
from pixlvault.db_models import Face
from tests.utils import upload_pictures_and_wait


def make_image(color=(0, 0, 0)):
    img = Image.new("RGB", (64, 64), color)
    buf = BytesIO()
    img.save(buf, format="PNG")
    return buf.getvalue()


def test_many_to_many_face_data():
    """
    Test that multiple characters can be associated with a picture, and that face data is stored per character.
    """
    with tempfile.TemporaryDirectory() as temp_dir:
        server_config_path = os.path.join(temp_dir, "server-config.json")
        with Server(server_config_path) as server:
            client = TestClient(server.api)

            resp = client.post(
                "/login", json={"username": "testuser", "password": "testpassword"}
            )
            assert resp.status_code == 200

            # Create two characters
            resp1 = client.post(
                "/characters", json={"name": "Alice", "description": "Test char 1"}
            )
            resp2 = client.post(
                "/characters", json={"name": "Bob", "description": "Test char 2"}
            )
            assert resp1.status_code == 200
            assert resp2.status_code == 200
            char1 = resp1.json()["character"]["id"]
            char2 = resp2.json()["character"]["id"]

            print(
                f"Created characters: Alice id={char1}, Bob id={char2}",
                file=sys.stdout,
                flush=True,
            )
            assert char1 is not None
            assert char2 is not None

            # Upload a picture
            img_bytes = make_image((255, 0, 0))
            files = [("file", ("red.png", img_bytes, "image/png"))]
            import_status = upload_pictures_and_wait(client, files)
            assert import_status["status"] == "completed"
            pic_id = import_status["results"][0]["picture_id"]

            # Simulate worker: insert two face rows for the picture
            server.vault.db.run_task(
                lambda session: (
                    session.add(
                        Face(picture_id=pic_id, face_index=0, bbox=[1, 2, 3, 4])
                    ),
                    session.commit(),
                )
            )
            server.vault.db.run_task(
                lambda session: (
                    session.add(
                        Face(picture_id=pic_id, face_index=1, bbox=[10, 20, 30, 40])
                    ),
                    session.commit(),
                )
            )

            # Now associate Alice and Bob with the faces by setting character_id
            server.vault.db.run_task(
                lambda session: (
                    session.query(Face)
                    .filter(Face.picture_id == pic_id, Face.face_index == 0)
                    .update({"character_id": int(char1)}),
                    session.commit(),
                )
            )
            server.vault.db.run_task(
                lambda session: (
                    session.query(Face)
                    .filter(Face.picture_id == pic_id, Face.face_index == 1)
                    .update({"character_id": int(char2)}),
                    session.commit(),
                )
            )

            # Retrieve and check face data for each character/face
            def get_face_data(session, picture_id, face_index):
                return Face.find(session, picture_id=picture_id, face_index=face_index)

            faces1 = server.vault.db.run_task(get_face_data, pic_id, 0)
            faces2 = server.vault.db.run_task(get_face_data, pic_id, 1)

            face1 = faces1[0] if len(faces1) > 0 else None
            face2 = faces2[0] if len(faces2) > 0 else None

            assert face1 is not None, "Face 1 not found"
            assert face2 is not None, "Face 2 not found"

            assert face1.bbox == [1, 2, 3, 4]
            assert face1.character_id == int(char1)
            assert face2.bbox == [10, 20, 30, 40]
            assert face2.character_id == int(char2)
            # Remove Bob from the picture by setting character_id to None and check association is gone
            server.vault.db.run_task(
                lambda session: (
                    session.query(Face)
                    .filter(Face.picture_id == pic_id, Face.face_index == 1)
                    .update({"character_id": None}),
                    session.commit(),
                )
            )
            face2_after = server.vault.db.run_task(
                lambda session: (
                    session.query(Face)
                    .filter(Face.picture_id == pic_id, Face.face_index == 1)
                    .first()
                )
            )
            assert face2_after.character_id is None, (
                f"Expected character_id None after removal, got: {face2_after}"
            )
    gc.collect()



import os
import tempfile
import numpy as np

from datetime import datetime
from io import BytesIO

from fastapi.testclient import TestClient
from PIL import Image
from sqlmodel import Session, select

from pixlvault.db_models.picture_set import PictureSet, PictureSetMember
from pixlvault.utils.image_processing.image_utils import ImageUtils
from pixlvault.server import Server


def _make_png_bytes(color: tuple[int, int, int]) -> bytes:
    img = Image.new("RGB", (24, 24), color=color)
    buf = BytesIO()
    img.save(buf, format="PNG")
    return buf.getvalue()


def test_picture_plugins_list_and_run_colour_filter():
    with tempfile.TemporaryDirectory() as temp_dir:
        server_config_path = os.path.join(temp_dir, "server_config.json")
        with Server(server_config_path=server_config_path) as server:
            client = TestClient(server.api)
            login_resp = client.post(
                "/login", json={"username": "testuser", "password": "testpassword"}
            )
            assert login_resp.status_code == 200

            img_a = _make_png_bytes((255, 40, 40))
            img_b = _make_png_bytes((40, 255, 40))

            def add_pictures(session: Session):
                first = ImageUtils.create_picture_from_bytes(
                    image_root_path=server.vault.image_root,
                    image_bytes=img_a,
                )
                second = ImageUtils.create_picture_from_bytes(
                    image_root_path=server.vault.image_root,
                    image_bytes=img_b,
                )
                now = datetime.utcnow()
                first.imported_at = now
                second.imported_at = now
                session.add(first)
                session.add(second)
                session.commit()
                session.refresh(first)
                session.refresh(second)
                picture_set = PictureSet(name="Plugin Test Set", description="test")
                session.add(picture_set)
                session.commit()
                session.refresh(picture_set)
                session.add(
                    PictureSetMember(set_id=picture_set.id, picture_id=first.id)
                )
                session.add(
                    PictureSetMember(set_id=picture_set.id, picture_id=second.id)
                )
                session.commit()
                return [first.id, second.id, picture_set.id]

            inserted = server.vault.db.run_task(add_pictures)
            inserted_ids = inserted[:2]
            created_set_id = inserted[2]
            assert len(inserted_ids) == 2

            pictures_resp = client.get("/pictures?fields=grid")
            assert pictures_resp.status_code == 200
            pictures = pictures_resp.json()
            assert pictures and len(pictures) >= 2
            selected_ids = sorted(inserted_ids)

            plugins_resp = client.get("/pictures/plugins")
            assert plugins_resp.status_code == 200
            plugins_payload = plugins_resp.json()
            plugins = plugins_payload.get("plugins") or []
            names = {plugin.get("name") for plugin in plugins}
            assert "colour_filter" in names
            assert "scaling" in names
            assert "brightness_contrast" in names
            assert "blur_sharpen" in names
            colour_schema = next(
                (plugin for plugin in plugins if plugin.get("name") == "colour_filter"),
                None,
            )
            assert colour_schema is not None
            assert colour_schema.get("supports_images") is True
            assert colour_schema.get("supports_videos") is True
            brightness_contrast_schema = next(
                (
                    plugin
                    for plugin in plugins
                    if plugin.get("name") == "brightness_contrast"
                ),
                None,
            )
            assert brightness_contrast_schema is not None
            assert brightness_contrast_schema.get("supports_images") is True
            assert brightness_contrast_schema.get("supports_videos") is True
            blur_sharpen_schema = next(
                (plugin for plugin in plugins if plugin.get("name") == "blur_sharpen"),
                None,
            )
            assert blur_sharpen_schema is not None
            assert blur_sharpen_schema.get("supports_images") is True
            assert blur_sharpen_schema.get("supports_videos") is True

            run_resp = client.post(
                "/pictures/plugins/colour_filter",
                json={
                    "picture_ids": selected_ids,
                    "parameters": {"mode": "sepia"},
                },
            )
            assert run_resp.status_code == 200, run_resp.text
            run_payload = run_resp.json()
            assert run_payload.get("status") == "success"
            created_ids = run_payload.get("created_picture_ids") or []
            assert len(created_ids) == 2

            def fetch_set_members(session: Session, set_id: int):
                members = session.exec(
                    select(PictureSetMember).where(PictureSetMember.set_id == set_id)
                ).all()
                return {int(member.picture_id) for member in members}

            set_member_ids = server.vault.db.run_task(fetch_set_members, created_set_id)
            for created_id in created_ids:
                assert int(created_id) in set_member_ids

            after_resp = client.get("/pictures?fields=grid")
            assert after_resp.status_code == 200
            after_pictures = {
                int(pic["id"]): pic
                for pic in after_resp.json()
                if pic.get("id") is not None
            }

            for source_id, created_id in zip(selected_ids, created_ids):
                source = after_pictures.get(int(source_id))
                created = after_pictures.get(int(created_id))
                assert source is not None
                assert created is not None
                assert source.get("stack_id") is not None
                assert created.get("stack_id") == source.get("stack_id")
                assert int(created.get("stack_position")) == 0

            scale_resp = client.post(
                "/pictures/plugins/scaling",
                json={
                    "picture_ids": selected_ids,
                    "parameters": {
                        "algorithm": "lanczos",
                        "scale_factor": "2.0",
                    },
                },
            )
            assert scale_resp.status_code == 200, scale_resp.text
            scale_payload = scale_resp.json()
            assert scale_payload.get("status") == "success"
            scaled_ids = scale_payload.get("created_picture_ids") or []
            assert len(scaled_ids) == 2

            scaled_after_resp = client.get("/pictures?fields=grid")
            assert scaled_after_resp.status_code == 200
            scaled_after_pictures = {
                int(pic["id"]): pic
                for pic in scaled_after_resp.json()
                if pic.get("id") is not None
            }
            for source_id, scaled_id in zip(selected_ids, scaled_ids):
                source = scaled_after_pictures.get(int(source_id))
                scaled = scaled_after_pictures.get(int(scaled_id))
                assert source is not None
                assert scaled is not None
                assert int(scaled.get("width")) == int(source.get("width")) * 2
                assert int(scaled.get("height")) == int(source.get("height")) * 2

            brightness_resp = client.post(
                "/pictures/plugins/brightness_contrast",
                json={
                    "picture_ids": selected_ids,
                    "parameters": {
                        "brightness": 1.1,
                        "contrast": 1.2,
                    },
                },
            )
            assert brightness_resp.status_code == 200, brightness_resp.text
            brightness_payload = brightness_resp.json()
            assert brightness_payload.get("status") == "success"
            brightness_ids = brightness_payload.get("created_picture_ids") or []
            assert len(brightness_ids) == 2

            blur_resp = client.post(
                "/pictures/plugins/blur_sharpen",
                json={
                    "picture_ids": selected_ids,
                    "parameters": {
                        "mode": "blur",
                        "strength": 1.0,
                    },
                },
            )
            assert blur_resp.status_code == 200, blur_resp.text
            blur_payload = blur_resp.json()
            assert blur_payload.get("status") == "success"
            blur_output_ids = blur_payload.get("output_picture_ids") or []
            assert len(blur_output_ids) == 2


def test_create_picture_from_bytes_preserves_video_extension_format(monkeypatch):
    class _FakeVideoCapture:
        def __init__(self, _path):
            self._read = False

        def read(self):
            if self._read:
                return False, None
            self._read = True
            frame = np.zeros((32, 48, 3), dtype=np.uint8)
            return True, frame

        def release(self):
            return None

    monkeypatch.setattr(
        "pixlvault.utils.image_processing.video_utils.cv2.VideoCapture",
        _FakeVideoCapture,
    )

    with tempfile.TemporaryDirectory() as temp_dir:
        picture = ImageUtils.create_picture_from_bytes(
            image_root_path=temp_dir,
            image_bytes=b"not-an-image",
            picture_uuid="example.webm",
        )

        assert picture.format == "WEBM"
        assert picture.file_path is not None
        assert picture.file_path.endswith(".webm")



from pixlvault.picture_tagger import PictureTagger
from pixlvault.tasks.missing_tag_finder import MissingTagFinder


def _build_tagger_for_budget_tests(
    budget_mb: int = 4096,
    expected_tasks: int = 2,
    max_concurrent: int = 64,
    onnx_capacity: int = 64,
    custom_batch: int = 16,
):
    tagger = PictureTagger.__new__(PictureTagger)
    tagger._device = "cuda"
    tagger._max_vram_usage_mb = budget_mb
    tagger._expected_concurrent_tag_tasks = expected_tasks
    tagger._onnx_batch_capacity = onnx_capacity
    tagger._custom_tagger_batch = custom_batch
    tagger.max_concurrent_images = lambda: max_concurrent
    return tagger


def test_vram_batch_cap_is_queue_aware():
    single_task = _build_tagger_for_budget_tests(expected_tasks=1)
    two_tasks = _build_tagger_for_budget_tests(expected_tasks=2)

    cap_single = single_task._vram_limited_batch_cap(base_mb=900, per_item_mb=220)
    cap_two = two_tasks._vram_limited_batch_cap(base_mb=900, per_item_mb=220)

    assert cap_two < cap_single
    assert cap_two <= 4


def test_estimated_task_vram_stays_within_budget_window():
    tagger = _build_tagger_for_budget_tests(
        budget_mb=4096,
        expected_tasks=2,
        max_concurrent=64,
        onnx_capacity=64,
        custom_batch=32,
    )

    estimate_mb = tagger.estimate_task_vram_mb(image_count=64)

    assert estimate_mb <= 4096
    assert estimate_mb >= 1200


def test_vram_cap_noop_on_cpu_mode():
    tagger = _build_tagger_for_budget_tests()
    tagger._device = "cpu"

    cap = tagger._vram_limited_batch_cap(base_mb=900, per_item_mb=220)

    assert cap == 10_000


def test_suggested_tag_task_size_tracks_effective_batch():
    tagger = _build_tagger_for_budget_tests(
        budget_mb=4096,
        expected_tasks=2,
        max_concurrent=64,
        onnx_capacity=64,
        custom_batch=32,
    )

    assert tagger._effective_wd14_batch_size() == 3
    assert tagger._effective_custom_batch_size() == 3
    assert tagger.suggested_tag_task_size() == 3


def test_custom_and_wd14_use_same_effective_batch_size():
    tagger = _build_tagger_for_budget_tests(
        budget_mb=4096,
        expected_tasks=2,
        max_concurrent=64,
        onnx_capacity=64,
        custom_batch=64,
    )

    assert tagger._effective_custom_batch_size() == tagger._effective_wd14_batch_size()


def test_incremental_vram_estimate_is_below_full_estimate():
    tagger = _build_tagger_for_budget_tests(
        budget_mb=4096,
        expected_tasks=2,
        max_concurrent=64,
        onnx_capacity=64,
        custom_batch=32,
    )

    full_estimate = tagger.estimate_task_vram_mb(image_count=64)
    incremental_estimate = tagger.estimate_task_incremental_vram_mb(image_count=64)

    assert incremental_estimate < full_estimate
    assert incremental_estimate >= 256


def test_missing_tags_finder_uses_suggested_task_size():
    class FakeTagger:
        def suggested_tag_task_size(self):
            return 3

        def max_concurrent_images(self):
            return 64

    class FakeDB:
        def __init__(self):
            self.image_root = "/tmp"

        def run_immediate_read_task(self, callback):
            class Picture:
                def __init__(self, pic_id):
                    self.id = pic_id

            return [Picture(i) for i in range(1, 30)]

    finder = MissingTagFinder(
        database=FakeDB(),
        picture_tagger_getter=lambda: FakeTagger(),
    )

    task = finder.find_task()

    assert task is not None
    assert task.params["batch_size"] == 3


def test_runtime_headroom_rounds_down_effective_batch_size():
    tagger = _build_tagger_for_budget_tests(
        budget_mb=8192,
        expected_tasks=2,
        max_concurrent=64,
        onnx_capacity=64,
        custom_batch=32,
    )

    tagger._query_process_vram_mb = lambda: 6404

    assert tagger._effective_wd14_batch_size() == 7
    assert tagger._effective_custom_batch_size() == 7
    assert tagger.suggested_tag_task_size() == 7



import threading
import time

from sqlmodel import Session

from pixlvault.database import DBPriority
from pixlvault.db_models import Picture
from pixlvault.picture_tagger import PictureTagger
from pixlvault.pixl_logging import get_logger
from pixlvault.tasks.base_task import BaseTask


logger = get_logger(__name__)


class DescriptionTask(BaseTask):
    """Task for generating and persisting description batches.

    Args:
        database: Vault database instance.
        picture_tagger: Tagger used to generate descriptions.
        pictures: Pictures to process in this batch.
    """

    CPU_SPILLOVER_REUSE_GRACE_S = 8.0
    _cpu_spillover_tagger: PictureTagger | None = None
    _cpu_spillover_last_used_at: float = 0.0
    _cpu_spillover_lock = threading.Lock()

    def __init__(
        self,
        database,
        picture_tagger: PictureTagger,
        pictures: list[Picture],
    ):
        picture_ids = [pic.id for pic in (pictures or []) if getattr(pic, "id", None)]
        super().__init__(
            task_type="DescriptionTask",
            params={
                "picture_ids": picture_ids,
                "batch_size": len(picture_ids),
            },
        )
        self._db = database
        self._picture_tagger = picture_tagger
        self._pictures = pictures or []
        self._cpu_spillover_enabled = False

    def allow_cpu_spillover(self) -> bool:
        return True

    def enable_cpu_spillover(self) -> None:
        self._cpu_spillover_enabled = True

    @classmethod
    def _acquire_cpu_spillover_tagger(cls, image_root: str) -> PictureTagger:
        with cls._cpu_spillover_lock:
            if cls._cpu_spillover_tagger is None:
                logger.debug("DescriptionTask: creating CPU spillover PictureTagger.")
                cls._cpu_spillover_tagger = PictureTagger(
                    silent=True,
                    device="cpu",
                    image_root=image_root,
                )
            cls._cpu_spillover_last_used_at = time.perf_counter()
            return cls._cpu_spillover_tagger

    @classmethod
    def _release_idle_cpu_spillover_tagger(cls, force: bool = False) -> None:
        with cls._cpu_spillover_lock:
            tagger = cls._cpu_spillover_tagger
            if tagger is None:
                return
            if not force:
                idle_s = time.perf_counter() - cls._cpu_spillover_last_used_at
                if idle_s < cls.CPU_SPILLOVER_REUSE_GRACE_S:
                    return
            cls._cpu_spillover_tagger = None
        try:
            tagger.close()
        except Exception as exc:
            logger.debug("DescriptionTask CPU spillover tagger close failed: %s", exc)

    def estimated_vram_mb(self) -> int:
        fn = getattr(self._picture_tagger, "estimate_description_vram_mb", None)
        if callable(fn):
            try:
                return max(0, int(fn(len(self._pictures))))
            except Exception:
                return 0
        return 0

    def _run_task(self):
        if not self._pictures:
            return {"changed_count": 0, "changed": []}

        descriptions_generated = self._generate_descriptions_batch(self._pictures)
        if not descriptions_generated:
            return {"changed_count": 0, "changed": []}

        def update_descriptions(session: Session, pics):
            changed = []
            for pic in pics:
                db_pic = session.get(Picture, pic.id)
                if db_pic is not None:
                    db_pic.description = pic.description
                    session.add(db_pic)
                    changed.append((Picture, pic.id, "description", pic.description))
            session.commit()
            return changed

        changed = self._db.run_task(
            update_descriptions,
            descriptions_generated,
            priority=DBPriority.LOW,
        )

        return {
            "changed_count": len(changed),
            "changed": changed,
        }

    def _generate_descriptions_batch(self, pictures: list[Picture]) -> list[Picture]:
        picture_ids = [pic.id for pic in pictures]
        logger.debug(
            "DescriptionTask: Generating descriptions for batch_size=%s ids=%s",
            len(pictures),
            picture_ids,
        )

        self._release_idle_cpu_spillover_tagger(force=False)
        active_tagger = self._picture_tagger
        cpu_spillover_tagger = None
        if self._cpu_spillover_enabled:
            logger.debug(
                "DescriptionTask %s: using CPU spillover for ids=%s",
                self.id,
                picture_ids,
            )
            cpu_spillover_tagger = self._acquire_cpu_spillover_tagger(
                self._db.image_root
            )
            active_tagger = cpu_spillover_tagger

        descriptions_generated = []
        try:
            batch_results = active_tagger.generate_descriptions_batch(pictures)
        except Exception as exc:
            import traceback

            logger.error(
                "DescriptionTask failed for ids=%s: %s\n%s",
                picture_ids,
                exc,
                traceback.format_exc(),
            )
            batch_results = None
        finally:
            if cpu_spillover_tagger is not None:
                with self._cpu_spillover_lock:
                    self._cpu_spillover_last_used_at = time.perf_counter()
                self._release_idle_cpu_spillover_tagger(force=False)

        if not batch_results:
            for pic in pictures:
                pic.description = ""
                descriptions_generated.append(pic)
            return descriptions_generated

        for pic in pictures:
            description = batch_results.get(pic.id)
            if description:
                pic.description = description
            else:
                logger.error("Failed to generate description for picture %s", pic.id)
                pic.description = ""
            descriptions_generated.append(pic)
        return descriptions_generated



"""Stack ordering utilities for picture stacks."""

from typing import List

from pixlvault.db_models import Picture
from pixlvault.utils.image_processing.image_utils import ImageUtils


class StackUtils:
    """Ordering helpers for picture stacks."""

    @staticmethod
    def picture_order_key(pic: Picture, image_root: str = None):
        """Return a sort key for a picture within a likeness stack.

        Ordering priority:
        - Higher resolution (width × height) first
        - Higher sharpness first
        - Lower noise_level first
        """
        if not pic.height or not pic.width:
            file_path = ImageUtils.resolve_picture_path(image_root, pic.file_path)
            pic.width, pic.height, _ = ImageUtils.load_metadata(file_path)
        resolution = (pic.width * pic.height) if pic.width and pic.height else 0

        quality = pic.quality
        sharp = quality.sharpness if quality and quality.sharpness is not None else 0.0
        noise = (
            quality.noise_level if quality and quality.noise_level is not None else 1.0
        )

        return (-resolution, -sharp, noise)

    @staticmethod
    def order_stack_pictures(
        pictures: List[Picture], image_root: str = None
    ) -> List[Picture]:
        """Return pictures sorted best-to-worst by resolution, sharpness, and noise."""
        return sorted(
            pictures, key=lambda pic: StackUtils.picture_order_key(pic, image_root)
        )



from pixlvault.db_models import Picture
from pixlvault.db_models.quality import Quality
from pixlvault.utils.stack.stack_utils import StackUtils


def test_order_stack_pictures_basic():
    pics = [
        Picture(
            file_path="a.jpg",
            width=100,
            height=100,
            quality=Quality(sharpness=0.5, noise_level=0.2),
        ),
        Picture(
            file_path="b.jpg",
            width=200,
            height=200,
            quality=Quality(sharpness=0.3, noise_level=0.1),
        ),
        Picture(
            file_path="c.jpg",
            width=100,
            height=100,
            quality=Quality(sharpness=0.9, noise_level=0.5),
        ),
    ]

    ordered = StackUtils.order_stack_pictures(pics)
    assert ordered[0].file_path == "b.jpg"
    assert ordered[1].file_path == "c.jpg"
    assert ordered[2].file_path == "a.jpg"


def test_order_stack_pictures_tiebreak():
    pics = [
        Picture(
            file_path="x.jpg",
            width=100,
            height=100,
            quality=Quality(sharpness=0.5, noise_level=0.2),
        ),
        Picture(
            file_path="y.jpg",
            width=100,
            height=100,
            quality=Quality(sharpness=0.5, noise_level=0.1),
        ),
        Picture(
            file_path="z.jpg",
            width=100,
            height=100,
            quality=Quality(sharpness=0.5, noise_level=0.3),
        ),
    ]
    ordered = StackUtils.order_stack_pictures(pics)
    assert ordered[0].file_path == "y.jpg"
    assert ordered[1].file_path == "x.jpg"
    assert ordered[2].file_path == "z.jpg"



import os
import time

from pixlvault.utils.image_processing.video_utils import VideoUtils

from .base_task_finder import BaseTaskFinder
from .watch_folder_import_task import (
    WatchFolderImportTask,
    load_watch_folders,
    persist_watch_folders,
)


class MissingWatchFolderImportFinder(BaseTaskFinder):
    """Find newly modified files in watch folders and create import tasks.

    Watch folder entries are read from the server config file as a list under
    the ``watch_folders`` key. Each entry must be a dict with the following
    fields:

        folder (str): Absolute path to the directory to monitor recursively.
        delete_after_import (bool): When True, source files are deleted from
            the watch folder after a successful import. Defaults to False.
        last_checked (float): Unix timestamp of the last scan. Updated
            automatically after each scan; do not set this manually.

    Example config entry::

        { "folder": "/home/user/downloads/photos", "delete_after_import": false }
    """

    _supported_image_exts = {
        ".jpg",
        ".jpeg",
        ".png",
        ".webp",
        ".bmp",
        ".heic",
        ".heif",
    }

    def __init__(self, database, config_path: str):
        super().__init__()
        self._db = database
        self._config_path = config_path

    def finder_name(self) -> str:
        return "MissingWatchFolderImportFinder"

    def find_task(self):
        watch_folders = load_watch_folders(self._config_path)
        if not watch_folders:
            return None

        now_ts = time.time()
        candidate_files = []
        total_candidates = 0
        updated = False

        for entry in watch_folders:
            folder = entry.get("folder")
            last_checked = float(entry.get("last_checked") or 0)
            delete_after_import = bool(entry.get("delete_after_import", False))

            if not folder or not os.path.isdir(folder):
                continue

            latest_seen = last_checked
            for root, _, files in os.walk(folder):
                for file_name in files:
                    file_path = os.path.join(root, file_name)
                    try:
                        mtime = os.path.getmtime(file_path)
                    except OSError:
                        continue
                    if not self._is_supported_file(file_path):
                        continue
                    if mtime > last_checked:
                        total_candidates += 1
                        candidate_files.append(
                            {
                                "file_path": file_path,
                                "delete_after_import": delete_after_import,
                            }
                        )
                    if mtime > latest_seen:
                        latest_seen = mtime

            entry["last_checked"] = max(latest_seen, now_ts)
            updated = True

        if updated and not candidate_files:
            persist_watch_folders(self._config_path, watch_folders)
            return None

        if not candidate_files:
            return None

        return WatchFolderImportTask(
            database=self._db,
            config_path=self._config_path,
            candidate_files=candidate_files,
            updated_watch_folders=watch_folders,
            total_candidates=total_candidates,
        )

    def _is_supported_file(self, file_path: str) -> bool:
        ext = os.path.splitext(file_path)[1].lower()
        if ext in self._supported_image_exts:
            return True
        return VideoUtils.is_video_file(file_path)



from sqlmodel import Column, ForeignKey, Integer, Relationship, SQLModel, Field

from typing import List, Optional, TYPE_CHECKING


if TYPE_CHECKING:
    from pixlvault.db_models.character import Character
    from pixlvault.db_models.picture import Picture


class PictureSetMember(SQLModel, table=True):
    """
    Database model for the picture_set_members table.
    Many-to-many junction table between picture_sets and pictures.
    """

    set_id: int = Field(
        default=None,
        sa_column=Column(
            Integer,
            ForeignKey("pictureset.id", ondelete="CASCADE"),
            primary_key=True,
            index=True,
        ),
    )
    picture_id: int = Field(
        default=None,
        sa_column=Column(
            Integer,
            ForeignKey("picture.id", ondelete="CASCADE"),
            primary_key=True,
            index=True,
        ),
    )


class PictureSet(SQLModel, table=True):
    """
    Database model for the picture_sets table.
    A picture set is a named collection of pictures.
    """

    id: int = Field(default=None, primary_key=True)
    name: str = Field(index=True, nullable=False)
    description: Optional[str] = Field(default=None)

    # Relationships
    members: List["Picture"] = Relationship(
        back_populates="picture_sets", link_model=PictureSetMember
    )
    reference_character: Optional["Character"] = Relationship(
        back_populates="reference_picture_set"
    )



from enum import auto, Enum


class EventType(Enum):
    CHANGED_PICTURES = auto()
    PICTURE_IMPORTED = auto()
    PLUGIN_PROGRESS = auto()
    CHANGED_TAGS = auto()
    CHANGED_CHARACTERS = auto()
    CHANGED_DESCRIPTIONS = auto()
    CHANGED_FACES = auto()
    QUALITY_UPDATED = auto()
    CLEARED_TAGS = auto()
